# -*- coding: utf-8 -*-
"""
Interface to pyrf random forest object detection.
"""
from __future__ import absolute_import, division, print_function
from os.path import exists, join
from ibeis.algo.detect import grabmodels
import utool as ut
import vtool_ibeis as vt
from six.moves import zip, map
import cv2
import random
(print, rrr, profile) = ut.inject2(__name__, '[randomforest]')

if not ut.get_argflag('--no-pyrf'):
    try:
        import pyrf
    except ImportError as ex:
        print('WARNING Failed to import pyrf. '
              'Randomforest detection is unavailable')
        if ut.SUPER_STRICT:
            raise

VERBOSE_RF = ut.get_argflag('--verbrf') or ut.VERBOSE


def train_gid_list(ibs, gid_list, trees_path=None, species=None, setup=True,
                   teardown=False, **kwargs):
    """
    Args:
        gid_list (list of int): the list of IBEIS image_rowids that need detection
        trees_path (str): the path that the trees will be saved into (along
            with temporary training inventory folders that are deleted once
            training is finished)
        species (str): the species that should be used to assign to the newly
            trained trees

    Kwargs (optional): refer to the PyRF documentation for configuration settings

    Returns:
        None
    """
    print("[randomforest.train()] training with %d gids and species=%r" % (
        len(gid_list), species, ))
    if trees_path is None and species is not None:
        trees_path = join(ibs.get_cachedir(), 'trees', species)

    # Get positive chip paths
    if species is None:
        aids_list = ibs.get_image_aids(gid_list)
    else:
        aids_list = ibs.get_image_aids_of_species(gid_list, species)

    # ##### TEMP #####
    # gid_list_ = []
    # aids_list_ = []
    # for gid, aid_list in zip(gid_list, aids_list):
    #     if len(aid_list) > 1:
    #         gid_list_.append(gid)
    #         aids_list_.append(aid_list)
    #     elif len(aid_list) == 1:
    #         (xtl, ytl, width, height) = ibs.get_annot_bboxes(aid_list)[0]
    #         if xtl > 5 and ytl > 5:
    #             gid_list_.append(gid)
    #             aids_list_.append(aid_list)
    # gid_list = gid_list_
    # aids_list = aids_list_
    # kwargs['trees_max_patches'] = 100000
    # ##### TEMP #####

    aid_list = ut.flatten(aids_list)
    train_pos_cpath_list = ibs.get_annot_chip_fpath(aid_list)

    # Ensure directories for negatives
    negatives_cache = join(ibs.get_cachedir(), 'pyrf_train_negatives')
    if (setup and not exists(negatives_cache)) or setup == 'force':
        # Force Check
        if exists(negatives_cache):
            ut.remove_dirs(negatives_cache)
        ut.ensuredir(negatives_cache)
        # Get negative chip paths
        print("[randomforest.train()] Mining %d negative patches" % (len(train_pos_cpath_list), ))
        train_neg_cpath_list = []
        while len(train_neg_cpath_list) < len(train_pos_cpath_list):
            sample = random.randint(0, len(gid_list) - 1)
            gid = gid_list[sample]
            img_width, img_height = ibs.get_image_sizes(gid)
            size = min(img_width, img_height)
            if species is None:
                aid_list = ibs.get_image_aids(gid)
            else:
                aid_list = ibs.get_image_aids_of_species(gid, species)
            annot_bbox_list = ibs.get_annot_bboxes(aid_list)
            # Find square patches
            square = random.randint(int(size / 4), int(size / 2))
            xmin = random.randint(0, img_width - square)
            xmax = xmin + square
            ymin = random.randint(0, img_height - square)
            ymax = ymin + square
            if _valid_candidate((xmin, xmax, ymin, ymax), annot_bbox_list):
                if VERBOSE_RF:
                    print("[%d / %d] MINING NEGATIVE PATCH (%04d, %04d, %04d, %04d) FROM GID %d" % (
                        len(train_neg_cpath_list), len(train_pos_cpath_list), xmin, xmax, ymin, ymax, gid, ))
                img = ibs.get_images(gid)
                img_path = join(negatives_cache, "neg_%07d.JPEG" % (len(train_neg_cpath_list), ))
                img = img[ymin:ymax, xmin:xmax]
                cv2.imwrite(img_path, img)
                train_neg_cpath_list.append(img_path)
    else:
        train_neg_cpath_list = ut.ls(negatives_cache, '*.JPEG')
        #direct = Directory(negatives_cache, include_extensions=['JPEG'])
        #train_neg_cpath_list = direct.files()

    # Train trees
    train_gpath_list(ibs, train_pos_cpath_list, train_neg_cpath_list,
                     trees_path=trees_path, species=species, **kwargs)

    # Remove cached negatives directory
    if teardown:
        ut.remove_dirs(negatives_cache)


def train_gpath_list(ibs, train_pos_cpath_list, train_neg_cpath_list,
                     trees_path=None, **kwargs):
    """
    Args:
        train_pos_cpath_list (list of str): the list of positive image paths
            for training
        train_neg_cpath_list (list of str): the list of negative image paths
            for training
        trees_path (str): the path that the trees will be saved into (along
            with temporary training inventory folders that are deleted once
            training is finished)
        species (str, optional): the species that should be used to assign to
            the newly trained trees

    Kwargs (optional): refer to the PyRF documentation for configuration settings

    Returns:
        None
    """
    if trees_path is None:
        trees_path = join(ibs.get_treesdir(), 'generic')
    # Train trees
    detector = pyrf.Random_Forest_Detector()
    detector.train(train_pos_cpath_list, train_neg_cpath_list, trees_path,
                   **kwargs)


def detect_gpath_list_with_species(ibs, gpath_list, species, **kwargs):
    """
    Args:
        gpath_list (list of str): the list of image paths that need detection
        species (str): the species that should be used to select the pre-trained
            random forest model
        downsample (bool, optional): a flag to indicate if the original image
            sizes should be used; defaults to True

            True:  ibs.get_image_detectpaths() is used
            False: ibs.get_image_paths() is used

    Kwargs (optional): refer to the PyRF documentation for configuration settings

    Yields:
        iter
    """
    tree_path_list = _get_models(ibs, species)
    results_iter = detect(ibs, gpath_list, tree_path_list, **kwargs)
    return results_iter


def detect_gid_list_with_species(ibs, gid_list, species, downsample=True,
                                 **kwargs):
    """
    Args:
        gid_list (list of int): the list of IBEIS image_rowids that need detection
        species (str): the species that should be used to select the pre-trained
            random forest model
        downsample (bool, optional): a flag to indicate if the original image
            sizes should be used; defaults to True

            True:  ibs.get_image_detectpaths() is used
            False: ibs.get_image_paths() is used

    Kwargs (optional): refer to the PyRF documentation for configuration settings

    Returns:
        iter

    CommandLine:
        python -m ibeis.algo.detect.randomforest --test-detect_gid_list_with_species

    Example:
        >>> # DISABLE_DOCTEST
        >>> from ibeis.algo.detect.randomforest import *  # NOQA
        >>> from ibeis.algo.detect.randomforest import _get_models  # NOQA
        >>> import ibeis
        >>> # build test data
        >>> ibs = ibeis.opendb('testdb1')
        >>> species = ibeis.const.TEST_SPECIES.ZEB_PLAIN
        >>> gid_list = ibs.get_valid_gids()
        >>> downsample = True
        >>> kwargs = {}
        >>> # execute function
        >>> result = detect_gid_list_with_species(ibs, gid_list, species, downsample)
        >>> # verify results
        >>> print(result)
    """
    tree_path_list = _get_models(ibs, species)
    results_iter = detect_gid_list(ibs, gid_list, tree_path_list,
                                   downsample=downsample, verbose=False,
                                   **kwargs)
    return results_iter


def detect_gid_list(ibs, gid_list, tree_path_list, downsample=True, **kwargs):
    """
    Args:
        gid_list (list of int): the list of IBEIS image_rowids that need detection
        tree_path_list (list of str): the list of trees to load for detection
        downsample (bool, optional): a flag to indicate if the original image
            sizes should be used; defaults to True

            True:  ibs.get_image_detectpaths() is used
            False: ibs.get_image_paths() is used

    Kwargs (optional): refer to the PyRF documentation for configuration settings

    Yields:
        results (list of dict)
    """
    # Get new gpaths if downsampling
    if downsample:
        gpath_list = ibs.get_image_detectpaths(gid_list)
        neww_list = [vt.open_image_size(gpath)[0] for gpath in gpath_list]
        oldw_list = [oldw for (oldw, oldh) in ibs.get_image_sizes(gid_list)]
        downsample_list = [oldw / neww for oldw, neww in zip(oldw_list, neww_list)]
    else:
        gpath_list = ibs.get_image_paths(gid_list)
        downsample_list = [None] * len(gpath_list)
    # Run detection
    results_iter = detect(ibs, gpath_list, tree_path_list, **kwargs)
    # Upscale the results
    for gid, downsample, (gpath, result_list) in zip(gid_list, downsample_list, results_iter):
        # Upscale the results back up to the original image size
        if downsample is not None and downsample != 1.0:
            for result in result_list:
                for key in ['centerx', 'centery', 'xtl', 'ytl', 'width', 'height']:
                    result[key] = int(result[key] * downsample)
        yield gid, gpath, result_list


def detect(ibs, gpath_list, tree_path_list, **kwargs):
    """
    Args:
        gpath_list (list of str): the list of image paths that need detection
        tree_path_list (list of str): the list of trees to load for detection

    Kwargs (optional): refer to the PyRF documentation for configuration settings

    Returns:
        iter
    """
    # Get scales from detect config, if not specified
    if 'scale_list' not in kwargs.keys():
        kwargs['scale_list'] = list(map(float, ibs.cfg.detect_cfg.scale_list.split(',')))
        assert all([ isinstance(scale, float) for scale in kwargs['scale_list'] ])

    verbose = kwargs.get('verbose', ut.VERBOSE)
    if verbose:
        print('[randomforest.detect()] Detecting with %d trees with scale_list=%r' % (
            len(tree_path_list), kwargs['scale_list'], ))

    # Run detection
    detector = pyrf.Random_Forest_Detector(verbose=verbose)
    forest = detector.forest(tree_path_list)
    results_iter = detector.detect(forest, gpath_list, **kwargs)
    return results_iter


########################


def _overlap_percentage(minmax_tup1, minmax_tup2):
    (xmin1, xmax1, ymin1, ymax1) = minmax_tup1
    (xmin2, xmax2, ymin2, ymax2) = minmax_tup2
    width1, height1 = xmax1 - xmin1, ymax1 - ymin1
    width2, height2 = xmax2 - xmin2, ymax2 - ymin2
    x_overlap = max(0, min(xmax1, xmax2) - max(xmin1, xmin2))
    y_overlap = max(0, min(ymax1, ymax2) - max(ymin1, ymin2))
    area_overlap = float(x_overlap * y_overlap)
    area_total = min(width1 * height1, width2 * height2)
    percentage = area_overlap / area_total
    return percentage


def _valid_candidate(candidate, annot_bbox_list, overlap=0.0, tries=10):
    for i in range(tries):
        valid = True
        for annot_bbox in annot_bbox_list:
            xtl, ytl, width, height = annot_bbox
            xmin, xmax, ymin, ymax = xtl, xtl + width, ytl, ytl + height
            if _overlap_percentage(candidate, (xmin, xmax, ymin, ymax)) > overlap:
                valid = False
                break  # break inner loop
        if valid:
            return True
    return False


def _get_models(ibs, species, modeldir='default', cfg_override=True, verbose=VERBOSE_RF):
    r"""
    Args:
        ibs (IBEISController):  ibeis controller object
        species (?):
        modeldir (str): (default = 'default')
        cfg_override (bool): (default = True)
        verbose (bool):  verbosity flag(default = False)

    Returns:
        ?: fpath_list

    CommandLine:
        python -m ibeis.algo.detect.randomforest --test-_get_models

    Example:
        >>> # ENABLE_DOCTEST
        >>> from ibeis.algo.detect.randomforest import *  # NOQA
        >>> import ibeis
        >>> ibs = ibeis.opendb(defaultdb='testdb1')
        >>> species = ibeis.const.TEST_SPECIES.ZEB_PLAIN
        >>> modeldir = 'default'
        >>> cfg_override = True
        >>> verbose = False
        >>> fpath_list = _get_models(ibs, species, modeldir, cfg_override, verbose)
        >>> result = ('fpath_list = %s' % (str(fpath_list),))
        >>> print(result)
    """
    #with ut.embed_on_exception_context:
    if cfg_override and len(ibs.cfg.detect_cfg.trees_path) > 0:
        trees_path = ibs.cfg.detect_cfg.trees_path
    else:
        # Ensure all models downloaded and accounted for
        assert species is not None, '[_get_models] Cannot detect without specifying a species'
        grabmodels.ensure_models(modeldir=modeldir, verbose=verbose)
        trees_path = grabmodels.get_species_trees_paths(species, modeldir=modeldir)
    # Load tree paths
    if ut.checkpath(trees_path, verbose=verbose):
        fpath_list = ut.ls(trees_path, '*.txt')
        #direct = Directory(trees_path, include_extensions=['txt'])
        #files = direct.files()
    else:
        # If the models do not exist, return None
        fpath_list = None
    if fpath_list is None or len(fpath_list) == 0:
        msg = ut.codeblock(
            '''
            [_get_models] Error loading trees, either directory or fpath_list not found
              * trees_path = %r
              * fpath_list = %r
              * species = %r
              * model_dir = %r
              * cfg_override = %r
            '''
        ) % (trees_path, fpath_list, species, modeldir, cfg_override)
        raise AssertionError(msg)
    return fpath_list



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
    This is a proof of concept for connecting to an authenticated Qubica Server
"""
from __future__ import print_function, division, absolute_import
from hashlib import sha1
import hmac
import requests

# System variables
APPLICATION_PROTOCOL   = 'http'
APPLICATION_DOMAIN     = '127.0.0.1'
APPLICATION_PORT       = None
APPLICATION_NAME       = 'IBEIS'
APPLICATION_SECRET_KEY = 'CB73808F-A6F6-094B-5FCD-385EBAFF8FC0'


def _raise(exception, message):
    raise exception('[%s] ERROR: %s' % (__file__, message))


def get_signature(key, message):
    return str(hmac.new(key, message, sha1).digest().encode("base64").rstrip('\n'))


def get_authorization_header(uri, user_email=None, user_enc_pass=None):
    # Get signature
    secret_key_signature = get_signature(APPLICATION_SECRET_KEY, uri)
    application_authentication = '%s:%s' % (APPLICATION_NAME, secret_key_signature, )
    if user_email is None or user_enc_pass is None:
        return '%s' % (application_authentication, )
    return '%s:%s:%s' % (application_authentication, user_email, user_enc_pass)


def _api_result(uri, method, user_email=None, user_enc_pass=None, **kwargs):
    """
        Make a general (method) API request to the server
    """
    # Make GET request to server
    method = method.upper()
    url = '%s://%s:%s%s' % (APPLICATION_PROTOCOL, APPLICATION_DOMAIN, APPLICATION_PORT, uri)
    header = get_authorization_header(url, user_email, user_enc_pass)
    headers = {'Authorization': header}
    args = (method, url, headers, kwargs, )
    print('Server request (%r): %r\n\tHeaders: %r\n\tArgs: %r' % args)
    try:
        if method == 'GET':
            req = requests.get(url, headers=headers, params=kwargs, verify=False)
        elif method == 'POST':
            req = requests.post(url, headers=headers, payload=kwargs, verify=False)
        else:
            _raise(KeyError, '_api_result got unsupported method=%r' % (method, ))
    except requests.exceptions.ConnectionError as ex:
        _raise(IOError, '_api_result could not connect to server %s' % (ex, ))
    return req.status_code, req.text, req.json


def get_api_result(uri, user_email=None, user_enc_pass=None, **kwargs):
    """
        Make a GET API request to the server
    """
    return _api_result(uri, 'get', user_email=user_email,
                       user_enc_pass=user_enc_pass, **kwargs)


def post_api_result(uri, user_email=None, user_enc_pass=None, **kwargs):
    """
        Make a GET API request to the server
    """
    return _api_result(uri, 'post', user_email=user_email,
                       user_enc_pass=user_enc_pass, **kwargs)


def run_test_api():
    r"""
    CommandLine:
        python -m ibeis.web.test_api --test-run_test_api

    Example:
        >>> # xdoctest: +REQUIRES(--web)
        >>> from ibeis.web.test_api import *  # NOQA
        >>> response = run_test_api()
        >>> print('Server response: %r' % (response, ))
        >>> result = response
        (200, u'{"status": {"cache": -1, "message": "", "code": 200, "success": true}, "response": "testdb1"}', <bound method Response.json of <Response [200]>>)
    """
    import ibeis
    import time
    global APPLICATION_PORT

    web_instance = ibeis.opendb_in_background(db='testdb1', web=True, precache=False)

    # Get the application port from the background process
    if APPLICATION_PORT is None:
        web_port = web_instance.get_web_port_via_scan()
        if web_port is None:
            raise ValueError('IA web server is not running on any expected port')
        APPLICATION_PORT = '%s' % (web_port, )
    assert APPLICATION_PORT is not None

    # let the webapi startup in the background
    time.sleep(.1)
    uri = '/api/core/dbname/'
    # Make GET request to the server as a test
    response = get_api_result(uri)
    status_code, text, json = response
    web_instance.terminate()
    return response


if __name__ == '__main__':
    """
    CommandLine:
        python -m ibeis.web.test_api
        python -m ibeis.web.test_api --allexamples
        python -m ibeis.web.test_api --allexamples --noface --nosrc
    """
    import multiprocessing
    multiprocessing.freeze_support()  # for win32
    import utool as ut  # NOQA
    ut.doctest_funcs()



# -*- coding: utf-8 -*-
"""
Dependencies: flask, tornado

SeeAlso:
    routes.turk_identification
"""
from __future__ import absolute_import, division, print_function, unicode_literals
from ibeis.control import controller_inject
from flask import url_for, request, current_app  # NOQA
import numpy as np   # NOQA
import utool as ut
import uuid
import requests
import six
ut.noinject('[apis_sync]')


CLASS_INJECT_KEY, register_ibs_method = (
    controller_inject.make_ibs_register_decorator(__name__))
register_api   = controller_inject.get_ibeis_flask_api(__name__)


REMOTE_TESTING = False


if REMOTE_TESTING:
    REMOTE_DOMAIN = '127.0.0.1'
    REMOTE_PORT = '5555'
    REMOTE_UUID = None
else:
    REMOTE_DOMAIN = '35.161.135.191'
    REMOTE_PORT = '5555'
    REMOTE_UUID = 'e468d14b-3a39-4165-8f62-16f9e3deea39'

    remote_args = ut.get_arg_dict()
    REMOTE_DOMAIN = remote_args.get('sync-domain', REMOTE_DOMAIN)
    REMOTE_PORT = remote_args.get('sync-port', REMOTE_PORT)
    REMOTE_UUID = remote_args.get('sync-uuid', REMOTE_UUID)

    if REMOTE_UUID in [True, '', 'none', 'None']:
        REMOTE_UUID = None


REMOTE_URL = 'http://%s:%s' % (REMOTE_DOMAIN, REMOTE_PORT, )
REMOTE_UUID = None if REMOTE_UUID is None else uuid.UUID(REMOTE_UUID)


def _construct_route_url(route_rule):
    if not route_rule.startswith('/'):
        route_rule = '/' + route_rule
    if not route_rule.endswith('/'):
        route_rule = route_rule + '/'
    route_url = '%s%s' % (REMOTE_URL, route_rule, )
    return route_url


def _verify_response(response):
    try:
        response_dict = ut.from_json(response.text)
    except ValueError:
        raise AssertionError('Could not get valid JSON response from server')
    status = response_dict.get('status', {})
    assert status.get('success', False)
    response = response_dict.get('response', None)
    return response


def _get(route_rule, **kwargs):
    route_url = _construct_route_url(route_rule)
    response = requests.get(route_url, **kwargs)
    return _verify_response(response)


def _assert_remote_online(ibs):
    try:
        version = _get('/api/core/db/version/')
        uuid = _get('/api/core/db/uuid/init/')
        assert version == ibs.get_database_version()
        if REMOTE_UUID is not None:
            assert uuid == REMOTE_UUID
    except:
        raise IOError('Remote IBEIS DETECT database offline at %s' % (REMOTE_URL, ))


@register_ibs_method
def _detect_remote_push_images(ibs, gid_list):
    route_url = _construct_route_url('/api/upload/image/')

    num_images = len(gid_list)
    image_path_list = ibs.get_image_paths(gid_list)
    for index, image_path in enumerate(image_path_list):
        print('\tSending %d / %d: %r' % (index, num_images, image_path, ))
        file_dict = {
            'image': open(image_path, 'rb'),
        }
        response = requests.post(route_url, files=file_dict)
        _verify_response(response)
        print('\t...sent')


@register_ibs_method
def _detect_remote_push_imageset(ibs, image_uuid_list):
    route_url = _construct_route_url('/api/image/imageset/text/json/')

    db_name = ibs.get_dbname()
    db_uuid = ibs.get_db_init_uuid()
    time_str = ut.get_timestamp()
    imageset_text = 'Sync from %s (%s) at %s' % (db_name, db_uuid, time_str)
    imageset_text_list = [imageset_text] * len(image_uuid_list)

    data_dict = {
        'image_uuid_list': image_uuid_list,
        'imageset_text_list': imageset_text_list,
    }
    for key in data_dict:
        data_dict[key] = ut.to_json(data_dict[key])
    response = requests.put(route_url, data=data_dict)
    _verify_response(response)


@register_ibs_method
def _detect_remote_push_annots(ibs, aid_list):
    route_url = _construct_route_url('/api/annot/json/')

    print('\tSending...')
    data_dict = {
        'image_uuid_list': ibs.get_annot_image_uuids(aid_list),
        'annot_uuid_list': ibs.get_annot_uuids(aid_list),
        'annot_bbox_list': ibs.get_annot_bboxes(aid_list),
    }
    for key in data_dict:
        data_dict[key] = ut.to_json(data_dict[key])
    response = requests.post(route_url, data=data_dict)
    _verify_response(response)
    print('\t...sent')


@register_ibs_method
def _detect_remote_push_metadata(ibs, route_rule, uuid_str, value_str,
                                 uuid_list, value_list):
    route_url = _construct_route_url(route_rule)

    print('\tSetting %s metadata for %s' % (route_rule, uuid_str, ))
    data_dict = {
        uuid_str: uuid_list,
        value_str: value_list,
    }
    for key in data_dict:
        data_dict[key] = ut.to_json(data_dict[key])
    response = requests.put(route_url, data=data_dict)
    _verify_response(response)
    print('\t...set')


@register_ibs_method
def _detect_remote_push_annot_metadata(ibs, annot_uuid_list):
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    ibs._detect_remote_push_metadata('/api/annot/bbox/json/',
                                     'annot_uuid_list',
                                     'bbox_list',
                                     annot_uuid_list,
                                     ibs.get_annot_bboxes(aid_list))
    ibs._detect_remote_push_metadata('/api/annot/theta/json/',
                                     'annot_uuid_list',
                                     'theta_list',
                                     annot_uuid_list,
                                     ibs.get_annot_thetas(aid_list))
    ibs._detect_remote_push_metadata('/api/annot/viewpoint/json/',
                                     'annot_uuid_list',
                                     'viewpoint_list',
                                     annot_uuid_list,
                                     ibs.get_annot_viewpoints(aid_list))
    ibs._detect_remote_push_metadata('/api/annot/quality/text/json/',
                                     'annot_uuid_list',
                                     'quality_text_list',
                                     annot_uuid_list,
                                     ibs.get_annot_quality_texts(aid_list))
    ibs._detect_remote_push_metadata('/api/annot/species/json/',
                                     'annot_uuid_list',
                                     'species_text_list',
                                     annot_uuid_list,
                                     ibs.get_annot_species_texts(aid_list))
    ibs._detect_remote_push_metadata('/api/annot/multiple/json/',
                                     'annot_uuid_list',
                                     'flag_list',
                                     annot_uuid_list,
                                     ibs.get_annot_multiple(aid_list))
    ibs._detect_remote_push_metadata('/api/annot/interest/json/',
                                     'annot_uuid_list',
                                     'flag_list',
                                     annot_uuid_list,
                                     ibs.get_annot_interest(aid_list))
    ibs._detect_remote_push_metadata('/api/annot/tags/json/',
                                     'annot_uuid_list',
                                     'annot_tags_list',
                                     annot_uuid_list,
                                     ibs.get_annot_tag_text(aid_list))
    ibs._detect_remote_push_metadata('/api/annot/name/text/json/',
                                     'annot_uuid_list',
                                     'name_text_list',
                                     annot_uuid_list,
                                     ibs.get_annot_name_texts(aid_list))


@register_ibs_method
def _detect_remote_push_parts(ibs, part_rowid_list):
    route_url = _construct_route_url('/api/part/json/')

    print('\tSending...')
    data_dict = {
        'annot_uuid_list': ibs.get_part_annot_uuids(part_rowid_list),
        'part_uuid_list': ibs.get_part_uuids(part_rowid_list),
        'part_bbox_list': ibs.get_part_bboxes(part_rowid_list),
    }
    for key in data_dict:
        data_dict[key] = ut.to_json(data_dict[key])
    response = requests.post(route_url, data=data_dict)
    _verify_response(response)
    print('\t...sent')


@register_ibs_method
def _detect_remote_push_part_metadata(ibs, part_uuid_list):
    part_rowid_list = ibs.get_part_rowids_from_uuid(part_uuid_list)
    ibs._detect_remote_push_metadata('/api/part/bbox/json/',
                                     'part_uuid_list',
                                     'bbox_list',
                                     part_uuid_list,
                                     ibs.get_part_bboxes(part_rowid_list))
    ibs._detect_remote_push_metadata('/api/part/theta/json/',
                                     'part_uuid_list',
                                     'theta_list',
                                     part_uuid_list,
                                     ibs.get_part_thetas(part_rowid_list))
    ibs._detect_remote_push_metadata('/api/part/viewpoint/json/',
                                     'part_uuid_list',
                                     'viewpoint_list',
                                     part_uuid_list,
                                     ibs.get_part_viewpoints(part_rowid_list))
    ibs._detect_remote_push_metadata('/api/part/quality/text/json/',
                                     'part_uuid_list',
                                     'quality_text_list',
                                     part_uuid_list,
                                     ibs.get_part_quality_texts(part_rowid_list))  # NOQA
    ibs._detect_remote_push_metadata('/api/part/type/json/',
                                     'part_uuid_list',
                                     'type_text_list',
                                     part_uuid_list,
                                     ibs.get_part_types(part_rowid_list))
    ibs._detect_remote_push_metadata('/api/part/tags/json/',
                                     'part_uuid_list',
                                     'part_tags_list',
                                     part_uuid_list,
                                     ibs.get_part_tag_text(part_rowid_list))


@register_ibs_method
@register_api('/api/sync/', methods=['GET'])
def detect_remote_sync_images(ibs, gid_list=None,
                              only_sync_missing_images=True):
    _assert_remote_online(ibs)

    if gid_list is None:
        gid_list = ibs.get_valid_gids()

    confirm_list = [
        ut.random_nonce()[:5]
        for _ in range(3)
    ]
    confirm_str = '-'.join(confirm_list)
    print('You are about to submit %d images to a remote DETECT database at %r with UUID=%r.' % (len(gid_list), REMOTE_URL, REMOTE_UUID, ))
    print('Only do this action if you are confident in the detection accuracy of the images, annotations, annotation metadata, parts and part metadata.')
    print('In order to continue, please type exactly the confirmation string %r' % (confirm_str, ))

    if six.PY2:
        input_func = raw_input
    else:
        input_func = input
    response_str = input_func('Confirmation string [Empty to abort]: ')
    response_str = response_str.lower()
    assert confirm_str == response_str, 'Confirmation string mismatch, aborting...'

    ############################################################################

    # Sync images
    image_uuid_list = ibs.get_image_uuids(gid_list)
    image_uuid_list_ = _get('/api/image/json/')

    missing_gid_list = [
        gid
        for gid, image_uuid in list(zip(gid_list, image_uuid_list))
        if image_uuid not in image_uuid_list_
    ]

    num_missing = len(missing_gid_list)
    if num_missing > 0:
        print('Need to push %d images...' % (num_missing, ))
        ibs._detect_remote_push_images(missing_gid_list)
        print('...pushed')

    # Filter only missing
    gid_list_ = missing_gid_list if only_sync_missing_images else gid_list
    image_uuid_list_ = ibs.get_image_uuids(gid_list_)

    ############################################################################

    # Sync imageset
    print('Setting imageset...')
    ibs._detect_remote_push_imageset(image_uuid_list_)
    print('...set')

    ############################################################################

    # Sync annots
    aid_list = ut.flatten(ibs.get_image_aids(gid_list_))
    annot_uuid_list = ibs.get_annot_uuids(aid_list)
    annot_uuid_list_ = _get('/api/annot/json/')

    missing_aid_list = [
        aid
        for aid, annot_uuid in list(zip(aid_list, annot_uuid_list))
        if annot_uuid not in annot_uuid_list_
    ]

    num_missing = len(missing_aid_list)
    if num_missing > 0:
        print('Need to push %d annots...' % (num_missing, ))
        ibs._detect_remote_push_annots(missing_aid_list)
        print('...pushed')

    ############################################################################

    # Sync annotation metadata
    print('Synching annotation metadata...')
    if len(annot_uuid_list) > 0:
        ibs._detect_remote_push_annot_metadata(annot_uuid_list)
    print('...synched')

    ############################################################################

    # Sync parts
    part_rowid_list = ut.flatten(ibs.get_annot_part_rowids(aid_list))
    part_uuid_list = ibs.get_part_uuids(part_rowid_list)
    part_uuid_list_ = _get('/api/part/json/')

    missing_part_rowid_list = [
        part_rowid
        for part_rowid, part_uuid in list(zip(part_rowid_list, part_uuid_list))
        if part_uuid not in part_uuid_list_
    ]

    num_missing = len(missing_part_rowid_list)
    if num_missing > 0:
        print('Need to push %d parts...' % (num_missing, ))
        ibs._detect_remote_push_parts(missing_part_rowid_list)
        print('...pushed')

    ############################################################################

    # Sync part metadata
    print('Synching part metadata...')
    if len(part_uuid_list) > 0:
        ibs._detect_remote_push_part_metadata(part_uuid_list)
    print('...synched')


if __name__ == '__main__':
    """
    CommandLine:
        python -m ibeis.web.app
        python -m ibeis.web.app --allexamples
        python -m ibeis.web.app --allexamples --noface --nosrc
    """
    import multiprocessing
    multiprocessing.freeze_support()  # for win32
    import utool as ut  # NOQA
    ut.doctest_funcs()



# -*- coding: utf-8 -*-
"""Interface to Azure object proposals."""
from __future__ import absolute_import, division, print_function
import requests
import utool as ut
import numpy as np
from six.moves import zip
from os.path import abspath, dirname, expanduser, join, exists  # NOQA
(print, rrr, profile) = ut.inject2(__name__, '[azure]')


VERBOSE_AZURE = ut.get_argflag('--verbazure') or ut.VERBOSE


NPROC_MULTIPLIER = 2


PREDICTION_URL = 'https://southcentralus.api.cognitive.microsoft.com/customvision/v2.0/Prediction/%s/image?iterationId=%s'
PREDICTION_HEADER = {
    'Prediction-Key': None,
    'Content-Type': 'application/octet-stream'
}
PREDICTION_DICT = {
    None  : ('9bb5790b-7f59-4c0b-b571-21e68d29f4b2', 'a4fb7280-b0be-4706-91c6-7651d116ac46', '34e5c511adfc449290e10868218906f9'),
}


def detect_gid_list(ibs, gid_list, verbose=VERBOSE_AZURE, **kwargs):
    """Detect gid_list with azure.

    Args:
        gid_list (list of int): the list of IBEIS image_rowids that need detection

    Kwargs (optional): refer to the Azure documentation for configuration settings

    Args:
        ibs (ibeis.IBEISController):  image analysis api
        gid_list (list of int): the list of IBEIS image_rowids that need detection

    Kwargs:
        detector, config_filepath, weights_filepath, verbose

    Yields:
        tuple: (gid, gpath, result_list)
    """
    # Get new gpaths if downsampling
    config = {
        'draw_annots': False,
        'thumbsize': 900,
    }
    gpath_list = ibs.get_image_thumbpath(gid_list, ensure_paths=True, **config)
    size_list = ibs.get_image_sizes(gid_list)

    # Run detection
    results_iter = detect(gpath_list, verbose=verbose, **kwargs)

    # Upscale the results
    _iter = zip(gid_list, size_list, results_iter)
    for gid, size, (gpath, result_list) in _iter:
        width, height = size

        # Upscale the results back up to the original image size
        for result in result_list:
            result['xtl']    = int(np.around(result['xtl']    * width ))
            result['ytl']    = int(np.around(result['ytl']    * height))
            result['width']  = int(np.around(result['width']  * width ))
            result['height'] = int(np.around(result['height'] * height))

        yield (gid, gpath, result_list)


def _detect(gpath, prediction_project, prediction_iteration, prediction_model):
    with open(gpath, 'rb') as image_file:
        data = image_file.read()

    prediction_url = PREDICTION_URL % (prediction_project, prediction_iteration, )
    prediction_header = PREDICTION_HEADER.copy()
    prediction_header['Prediction-Key'] = prediction_model
    response = requests.post(url=prediction_url, data=data, headers=prediction_header)

    response_json = response.json()
    output_list = response_json['predictions']

    return output_list


def detect(gpath_list, config_filepath, verbose=VERBOSE_AZURE, **kwargs):
    """Detect image filepaths with azure.

    Args:
        gpath_list (list of str): the list of image paths that need proposal candidates

    Kwargs (optional): refer to the Azure documentation for configuration settings

    Returns:
        iter
    """
    # Get correct weight if specified with shorthand
    if config_filepath not in PREDICTION_DICT:
        config_filepath = None

    prediction = PREDICTION_DICT.get(config_filepath, None)
    assert prediction is not None, 'Azure needs to have a model configuration'
    prediction_project, prediction_iteration, prediction_model = prediction

    prediction_project_list   = [prediction_project] * len(gpath_list)
    prediction_iteration_list = [prediction_iteration] * len(gpath_list)
    prediction_model_list     = [prediction_model]   * len(gpath_list)
    arg_iter = list(zip(gpath_list, prediction_project_list, prediction_iteration_list, prediction_model_list))
    nprocs = ut.util_parallel.get_default_numprocs()
    nprocs *= NPROC_MULTIPLIER
    nprocs = min(nprocs, len(arg_iter))
    outputs_list = ut.util_parallel.generate2(_detect, arg_iter, nprocs=nprocs, ordered=True)

    # Execute detector for each image
    results_list_ = []
    for output_list in outputs_list:
        result_list_ = []
        for output in list(output_list):
            result_dict = {
                'xtl'        : output['boundingBox']['left'],
                'ytl'        : output['boundingBox']['top'],
                'width'      : output['boundingBox']['width'],
                'height'     : output['boundingBox']['height'],
                'class'      : output['tagName'],
                'confidence' : output['probability'],
            }
            result_list_.append(result_dict)
        results_list_.append(result_list_)

    if len(results_list_) != len(gpath_list):
        raise ValueError('Azure did not return valid data')

    results_list = zip(gpath_list, results_list_)
    return results_list



# -*- coding: utf-8 -*-
"""
"""
from __future__ import absolute_import, division, print_function, unicode_literals
import utool as ut
(print, print_, printDBG, rrr, profile) = ut.inject(__name__, '[preproc_residual]')


def add_residual_params_gen(ibs, fid_list, qreq_=None):
    return None


def on_delete(ibs, featweight_rowid_list):
    print('Warning: Not Implemented')
    print('Probably nothing to do here')


if __name__ == '__main__':
    import multiprocessing
    multiprocessing.freeze_support()
    testable_list = [
    ]
    ut.doctest_funcs(testable_list)



# -*- coding: utf-8 -*-
"""
This module defines the entry point into the IBEIS system
ibeis.opendb and ibeis.main are the main entry points
"""
from __future__ import absolute_import, division, print_function
#from six.moves import builtins
import sys
import multiprocessing

#try:
import utool as ut
profile = ut.profile
#profile = getattr(builtins, 'profile')
#except AttributeError:
#def profile(func):
#    return func

QUIET = '--quiet' in sys.argv
NOT_QUIET = not QUIET
USE_GUI = '--gui' in sys.argv or '--nogui' not in sys.argv


def _on_ctrl_c(signal, frame):
    proc_name = multiprocessing.current_process().name
    print('[ibeis.main_module] Caught ctrl+c in %s' % (proc_name,))
    sys.exit(0)
    # try:
    #     _close_parallel()
    # except Exception as ex:
    #     print('Something very bad happened' + repr(ex))
    # finally:
    #     print('[ibeis.main_module] sys.exit(0)')
    #     sys.exit(0)

#-----------------------
# private init functions


def _init_signals():
    import signal
    signal.signal(signal.SIGINT, _on_ctrl_c)


def _reset_signals():
    import signal
    signal.signal(signal.SIGINT, signal.SIG_DFL)  # reset ctrl+c behavior


def _parse_args():
    from ibeis import params
    params.parse_args()


def _init_matplotlib():
    from plottool_ibeis import __MPL_INIT__
    __MPL_INIT__.init_matplotlib()


def _init_gui(activate=True):
    import guitool_ibeis
    if NOT_QUIET:
        print('[main] _init_gui()')
    guitool_ibeis.ensure_qtapp()
    #USE_OLD_BACKEND = '--old-backend' in sys.argv
    #if USE_OLD_BACKEND:
    from ibeis.gui import guiback
    back = guiback.MainWindowBackend()
    #else:
    #    from ibeis.gui import newgui
    #    back = newgui.IBEISGuiWidget()
    if activate:
        guitool_ibeis.activate_qwindow(back.mainwin)
    return back


def _init_ibeis(dbdir=None, verbose=None, use_cache=True, web=None, **kwargs):
    """
    Private function that calls code to create an ibeis controller
    """
    import utool as ut
    from ibeis import params
    from ibeis.control import IBEISControl
    if verbose is None:
        verbose = ut.VERBOSE
    if verbose and NOT_QUIET:
        print('[main] _init_ibeis()')
    # Use command line dbdir unless user specifies it
    if dbdir is None:
        ibs = None
        print('[main!] WARNING: args.dbdir is None')
    else:
        kwargs = kwargs.copy()
        request_dbversion = kwargs.pop('request_dbversion', None)
        force_serial = kwargs.get('force_serial', None)
        ibs = IBEISControl.request_IBEISController(
            dbdir=dbdir, use_cache=use_cache,
            request_dbversion=request_dbversion,
            force_serial=force_serial)
        if web is None:
            web = ut.get_argflag(('--webapp', '--webapi', '--web', '--browser'),
                                 help_='automatically launch the web app / web api')
            #web = params.args.webapp
        if web:
            from ibeis.web import app
            port = params.args.webport
            app.start_from_ibeis(ibs, port=port, **kwargs)
    return ibs


def _init_parallel():
    import utool as ut
    if ut.VERBOSE:
        print('_init_parallel')
    from utool import util_parallel
    from ibeis import params
    # Import any modules which parallel process will use here
    # so they are accessable when the program forks
    #from utool import util_sysreq
    #util_sysreq.ensure_in_pythonpath('hesaff')
    #util_sysreq.ensure_in_pythonpath('pyrf')
    #util_sysreq.ensure_in_pythonpath('code')
    #import pyhesaff  # NOQA
    #import pyrf  # NOQA
    from ibeis import core_annots  # NOQA
    #.algo.preproc import preproc_chip  # NOQA
    util_parallel.set_num_procs(params.args.num_procs)
    #if PREINIT_MULTIPROCESSING_POOLS:
    #    util_parallel.init_pool(params.args.num_procs)


# def _close_parallel():
#     #if ut.VERBOSE:
#     #    print('_close_parallel')
#     try:
#         from utool import util_parallel
#         util_parallel.close_pool(terminate=True)
#     except Exception as ex:
#         import utool as ut
#         ut.printex(ex, 'error closing parallel')
#         raise


def _init_numpy():
    import utool as ut
    import numpy as np
    if ut.VERBOSE:
        print('_init_numpy')
    error_options = ['ignore', 'warn', 'raise', 'call', 'print', 'log']
    on_err = error_options[0]
    #np.seterr(divide='ignore', invalid='ignore')
    numpy_err = {
        'divide':  on_err,
        'over':    on_err,
        'under':   on_err,
        'invalid': on_err,
    }
    #numpy_print = {
    #    'precision': 8,
    #    'threshold': 500,
    #    'edgeitems': 3,
    #    'linewidth': 200,  # default 75
    #    'suppress': False,
    #    'nanstr': 'nan',
    #    'formatter': None,
    #}
    np.seterr(**numpy_err)
    #np.set_printoptions(**numpy_print)


#-----------------------
# private loop functions


def _guitool_loop(main_locals, ipy=False):
    import guitool_ibeis
    from ibeis import params
    print('[main] guitool_ibeis loop')
    back = main_locals.get('back', None)
    if back is not None:
        loop_freq = params.args.loop_freq
        ipy = ipy or params.args.cmd
        guitool_ibeis.qtapp_loop(qwin=back.mainwin, ipy=ipy, frequency=loop_freq, init_signals=False)
        if ipy:  # If we're in IPython, the qtapp loop won't block, so we need to refresh
            back.refresh_state()
    else:
        if NOT_QUIET:
            print('WARNING: back was not expected to be None')


def set_newfile_permissions():
    r"""
    sets this processes default permission bits when creating new files

    CommandLine:
        python -m ibeis.main_module --test-set_newfile_permissions

    Example:
        >>> # ENABLE_DOCTEST
        >>> from ibeis.main_module import *  # NOQA
        >>> import os
        >>> import utool as ut
        >>> # write before umask
        >>> ut.delete('tempfile1.txt')
        >>> ut.write_to('tempfile1.txt', 'foo')
        >>> stat_result1 = os.stat('tempfile1.txt')
        >>> # apply umask
        >>> set_newfile_permissions()
        >>> ut.delete('tempfile2.txt')
        >>> ut.write_to('tempfile2.txt', 'foo')
        >>> stat_result2 = os.stat('tempfile2.txt')
        >>> # verify results
        >>> print('old masked all bits = %o' % (stat_result1.st_mode))
        >>> print('new masked all bits = %o' % (stat_result2.st_mode))
    """
    import os
    #import stat
    # Set umask so all files written will be group read and writable
    # To get the permissions we want subtract what you want from 0o0666 because
    # umask subtracts the mask you give it.
    #mask = stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP | stat.S_IWGRP | stat.S_IROTH
    #mask = 0o000  # most permissive umask
    mask = 0o000  # most permissive umask
    prev_mask = os.umask(mask)
    return prev_mask
    #print('prev_mask = %o' % (prev_mask,))
    #print('new_mask  = %o' % (mask,))


def main(gui=True, dbdir=None, defaultdb='cache',
         allow_newdir=False, db=None,
         delete_ibsdir=False,
         **kwargs):
    """
    Program entry point
    Inits the system environment, an IBEISControl, and a GUI if requested

    Args:
        gui (bool): (default=True) If gui is False a gui instance will not be created
        dbdir (None): full directory of a database to load
        db (None): name of database to load relative to the workdir
        allow_newdir (bool): (default=False) if False an error is raised if a
            a new database is created
        defaultdb (str): codename of database to load if db and dbdir is None. a value
            of 'cache' will open the last database opened with the GUI.

    Returns:
        dict: main_locals
    """
    set_newfile_permissions()
    from ibeis.init import main_commands
    from ibeis.init import sysres
    # Display a visible intro message
    msg = '''
    _____ ______  _______ _____ _______
      |   |_____] |______   |   |______
    __|__ |_____] |______ __|__ ______|
    '''
    if NOT_QUIET:
        print(msg)
    # Init the only two main system api handles
    ibs = None
    back = None
    if NOT_QUIET:
        print('[main] ibeis.main_module.main()')
    _preload()
    DIAGNOSTICS = NOT_QUIET
    if DIAGNOSTICS:
        import os
        import utool as ut
        import ibeis
        print('[main] MAIN DIAGNOSTICS')
        print('[main]  * username = %r' % (ut.get_user_name()))
        print('[main]  * ibeis.__version__ = %r' % (ibeis.__version__,))
        print('[main]  * computername = %r' % (ut.get_computer_name()))
        print('[main]  * cwd = %r' % (os.getcwd(),))
        print('[main]  * sys.argv = %r' % (sys.argv,))
    # Parse directory to be loaded from command line args
    # and explicit kwargs
    if defaultdb in ['testdb1', 'testdb0']:
        from ibeis.tests.reset_testdbs import ensure_smaller_testingdbs
        ensure_smaller_testingdbs()
        #
    dbdir = sysres.get_args_dbdir(defaultdb=defaultdb,
                                  allow_newdir=allow_newdir, db=db,
                                  dbdir=dbdir)
    if delete_ibsdir is True:
        from ibeis.other import ibsfuncs
        assert allow_newdir, 'must be making new directory if you are deleting everything!'
        ibsfuncs.delete_ibeis_database(dbdir)

    #limit = sys.getrecursionlimit()
    #if limit == 1000:
    #    print('Setting Recursion Limit to 3000')
    #    sys.setrecursionlimit(3000)
    # Execute preload commands
    main_commands.preload_commands(dbdir, **kwargs)  # PRELOAD CMDS
    try:
        # Build IBEIS Control object
        ibs = _init_ibeis(dbdir)
        if gui and USE_GUI:
            back = _init_gui(activate=kwargs.get('activate', True))
            back.connect_ibeis_control(ibs)
    except Exception as ex:
        print('[main()] IBEIS LOAD encountered exception: %s %s' % (type(ex), ex))
        raise
    main_commands.postload_commands(ibs, back)  # POSTLOAD CMDS
    main_locals = {'ibs': ibs, 'back': back}
    return main_locals


def opendb_in_background(*args, **kwargs):
    """
    Starts a web server in the background
    """
    import utool as ut
    import time
    sec = kwargs.pop('wait', 0)
    if sec != 0:
        raise AssertionError('wait is depricated')
        print('waiting %s seconds for startup' % (sec,))
    proc = ut.spawn_background_process(opendb, *args, **kwargs)
    if sec != 0:
        raise AssertionError('wait is depricated')
        time.sleep(sec)  # wait for process to initialize
    return proc


def opendb_bg_web(*args, **kwargs):
    """
    Wrapper around opendb_in_background, returns a nice web_ibs
    object to execute web calls using normal python-like syntax

    Args:
        *args: passed to opendb_in_background
        **kwargs:
            port (int):
            domain (str): if specified assumes server is already running
                somewhere otherwise kwargs is passed to opendb_in_background
            start_job_queue (bool)

    Returns:
        web_ibs - this is a KillableProcess object with special functions

    CommandLine:
        python -m ibeis.main_module opendb_bg_web

    Example:
        >>> # DISABLE_DOCTEST
        >>> from ibeis.main_module import *  # NOQA
        >>> args = tuple()
        >>> kwargs = {}
        >>> print('Opening a web_ibs')
        >>> web_ibs = opendb_bg_web()
        >>> print('SUCESS Opened a web_ibs!')
        >>> print(web_ibs)
        >>> print('Now kill the web_ibs')
        >>> web_ibs.terminate2()
    """
    import utool as ut
    from ibeis.web import appfuncs
    domain = kwargs.pop('domain', ut.get_argval('--domain', type_=str, default=None))
    port = kwargs.pop('port', appfuncs.DEFAULT_WEB_API_PORT)

    if 'wait' in kwargs:
        print('NOTE: No need to specify wait param anymore. '
              'This is automatically taken care of.')

    if domain is None:
        # Requesting a local test server
        _kw = dict(web=True, browser=False)
        _kw.update(kwargs)
        web_ibs = opendb_in_background(*args, **_kw)
    else:
        # Using a remote controller, no need to spin up anything
        web_ibs = ut.DynStruct()
        web_ibs.terminate2 = lambda: None
    # Augment web instance with usefull test functions
    if domain is None:
        domain = 'http://127.0.1.1'
    if not domain.startswith('http://'):
        domain = 'http://' + domain
    baseurl = domain  + ':' + str(port)

    web_ibs.domain = domain
    web_ibs.port = port
    web_ibs.baseurl = baseurl

    def get(suffix, **kwargs):
        import requests
        return requests.get(baseurl + suffix)

    def post(suffix, **kwargs):
        import requests
        return requests.post(baseurl + suffix)

    def send_ibeis_request(suffix, type_='post', **kwargs):
        """
        Posts a request to a url suffix
        """
        import requests
        import utool as ut
        if not suffix.endswith('/'):
            raise Exception('YOU PROBABLY WANT A / AT THE END OF YOUR URL')
        payload = ut.map_dict_vals(ut.to_json, kwargs)
        if type_ == 'post':
            resp = requests.post(baseurl + suffix, data=payload)
            json_content = resp._content
        elif type_ == 'get':
            resp = requests.get(baseurl + suffix, data=payload)
            json_content = resp.content
        try:
            content = ut.from_json(json_content)
        except ValueError:
            raise Exception('Expected JSON string but got json_content=%r' % (json_content,))
        else:
            # print('content = %r' % (content,))
            if content['status']['code'] != 200:
                print(content['status']['message'])
                raise Exception(content['status']['message'])
        request_response = content['response']
        return request_response

    def wait_for_results(jobid, timeout=None, delays=[1, 3, 10]):
        """
        Waits for results from an engine
        """
        for _ in ut.delayed_retry_gen(delays):
            print('Waiting for jobid = %s' % (jobid,))
            status_response = web_ibs.send_ibeis_request('/api/engine/job/status/', jobid=jobid)
            if status_response['jobstatus'] == 'completed':
                break
        return status_response

    def read_engine_results(jobid):
        result_response = web_ibs.send_ibeis_request('/api/engine/job/result/', jobid=jobid)
        return result_response

    def send_request_and_wait(suffix, type_='post', timeout=None, **kwargs):
        jobid = web_ibs.send_ibeis_request(suffix, type_=type_, **kwargs)
        status_response = web_ibs.wait_for_results(jobid, timeout)  # NOQA
        result_response = web_ibs.read_engine_results(jobid)
        #>>> cmdict = ut.from_json(result_response['json_result'])[0]
        return result_response

    web_ibs.send_ibeis_request = send_ibeis_request
    web_ibs.wait_for_results = wait_for_results
    web_ibs.read_engine_results = read_engine_results
    web_ibs.send_request_and_wait = send_request_and_wait
    web_ibs.get = get
    web_ibs.post = post

    def wait_until_started():
        """ waits until the web server responds to a request """
        import requests
        for count in ut.delayed_retry_gen([1], timeout=15):
            if True or ut.VERBOSE:
                print('Waiting for server to be up. count=%r' % (count,))
            try:
                web_ibs.send_ibeis_request('/api/test/heartbeat/', type_='get')
                break
            except requests.ConnectionError:
                pass
    wait_until_started()
    return web_ibs


def opendb_fg_web(*args, **kwargs):
    """
    Ignore:
        >>> from ibeis.main_module import *  # NOQA
        >>> kwargs = {'db': 'testdb1'}
        >>> args = tuple()

        >>> import ibeis
        >>> ibs = ibeis.opendb_fg_web()

    """
    # Gives you context inside the web app for testing
    kwargs['start_web_loop'] = False
    kwargs['web'] = True
    kwargs['browser'] = False
    ibs = opendb(*args, **kwargs)
    from ibeis.control import controller_inject
    app = controller_inject.get_flask_app()
    ibs.app = app
    return ibs


def opendb(db=None, dbdir=None, defaultdb='cache', allow_newdir=False,
           delete_ibsdir=False, verbose=False, use_cache=True,
           web=None, **kwargs):
    """
    main without the preload (except for option to delete database before
    opening)

    Args:
        db (str):  database name in your workdir used only if dbdir is None
        dbdir (None): full database path
        defaultdb (str): dbdir search stratagy when db is None and dbdir is
            None
        allow_newdir (bool): (default=True) if True errors when opening a
            nonexisting database
        delete_ibsdir (bool): BE CAREFUL! (default=False) if True deletes the
            entire
        verbose (bool): verbosity flag
        web (bool): starts webserver if True (default=param specification)
        use_cache (bool): if True will try to return a previously loaded
            controller

    Returns:
        ibeis.IBEISController: ibs

    Example:
        >>> # ENABLE_DOCTEST
        >>> from ibeis.main_module import *  # NOQA
        >>> db = None
        >>> dbdir = None
        >>> defaultdb = 'cache'
        >>> allow_newdir = False
        >>> delete_ibsdir = False
        >>> verbose = False
        >>> use_cache = True
        >>> ibs = opendb(db, dbdir, defaultdb, allow_newdir, delete_ibsdir,
        >>>              verbose, use_cache)
        >>> result = str(ibs)
        >>> print(result)
    """
    from ibeis.init import sysres
    from ibeis.other import ibsfuncs
    dbdir = sysres.get_args_dbdir(defaultdb=defaultdb,
                                  allow_newdir=allow_newdir, db=db,
                                  dbdir=dbdir)
    if delete_ibsdir is True:
        assert allow_newdir, (
            'must be making new directory if you are deleting everything!')
        ibsfuncs.delete_ibeis_database(dbdir)
    ibs = _init_ibeis(dbdir, verbose=verbose, use_cache=use_cache, web=web,
                      **kwargs)
    return ibs


def start(*args, **kwargs):
    """ alias for main() """  # + main.__doc__
    return main(*args, **kwargs)


def opendb_test(gui=True, dbdir=None, defaultdb='cache', allow_newdir=False,
                db=None):
    """ alias for main() """  # + main.__doc__
    from ibeis.init import sysres
    _preload()
    dbdir = sysres.get_args_dbdir(defaultdb=defaultdb,
                                  allow_newdir=allow_newdir, db=db,
                                  dbdir=dbdir)
    ibs = _init_ibeis(dbdir)
    return ibs


def _preload(mpl=True, par=True, logging=True):
    """ Sets up python environment """
    import utool as ut
    #from ibeis.init import main_helpers
    from ibeis import params
    if  multiprocessing.current_process().name != 'MainProcess':
        return
    if ut.VERBOSE:
        print('[ibeis] _preload')
    _parse_args()
    # mpl backends
    if logging and not params.args.nologging:
        # Log in the configured ibeis log dir (which is maintained by utool)
        # fix this to be easier to figure out where the logs actually are
        ut.start_logging(appname='ibeis')
    if mpl:
        _init_matplotlib()
    # numpy print settings
    _init_numpy()
    # parallel servent processes
    if par:
        _init_parallel()
    # ctrl+c
    _init_signals()
    # inject colored exceptions
    ut.util_inject.inject_colored_exceptions()
    # register type aliases for debugging
    #main_helpers.register_utool_aliases()
    #return params.args


def main_loop(main_locals, rungui=True, ipy=False, persist=True):
    """
    Runs the qt loop if the GUI was initialized and returns an executable string
    for embedding an IPython terminal if requested.

    If rungui is False the gui will not loop even if back has been created

    the main locals dict must be callsed main_locals in the scope you call this
    function in.

    Args:
        main_locals (dict_):
        rungui      (bool):
        ipy         (bool):
        persist     (bool):

    Returns:
        str: execstr
    """
    print('[main] ibeis.main_module.main_loop()')
    from ibeis import params
    import utool as ut
    #print('current process = %r' % (multiprocessing.current_process().name,))
    #== 'MainProcess':
    if rungui and not params.args.nogui:
        try:
            _guitool_loop(main_locals, ipy=ipy)
        except Exception as ex:
            ut.printex(ex, 'error in main_loop')
            raise
    #if not persist or params.args.cmd:
    #    main_close()
    # Put locals in the exec namespace
    ipycmd_execstr = ut.ipython_execstr()
    locals_execstr = ut.execstr_dict(main_locals, 'main_locals')
    execstr = locals_execstr + '\n' + ipycmd_execstr
    return execstr


def main_close(main_locals=None):
    #import utool as ut
    #if ut.VERBOSE:
    #    print('main_close')
    # _close_parallel()
    _reset_signals()


#if __name__ == '__main__':
#    multiprocessing.freeze_support()
if __name__ == '__main__':
    """
    CommandLine:
        python -m ibeis.main_module
        python -m ibeis.main_module --allexamples
        python -m ibeis.main_module --allexamples --noface --nosrc
    """
    multiprocessing.freeze_support()  # for win32
    import utool as ut  # NOQA
    ut.doctest_funcs()



# -*- coding: utf-8 -*-
"""Dependencies: flask, tornado."""
from __future__ import absolute_import, division, print_function
from ibeis.control import accessor_decors, controller_inject
from ibeis import constants as const
import utool as ut
import simplejson as json
from os.path import join, dirname, abspath
from flask import url_for, request, current_app
from ibeis.constants import KEY_DEFAULTS, SPECIES_KEY
from ibeis.web import appfuncs as appf


USE_LOCALIZATIONS = True


CLASS_INJECT_KEY, register_ibs_method = (
    controller_inject.make_ibs_register_decorator(__name__))
register_api   = controller_inject.get_ibeis_flask_api(__name__)
register_route = controller_inject.get_ibeis_flask_route(__name__)


@register_ibs_method
@accessor_decors.default_decorator
@accessor_decors.getter_1to1
@register_api('/api/detect/randomforest/', methods=['PUT', 'GET'])
def detect_random_forest(ibs, gid_list, species, commit=True, **kwargs):
    """Run animal detection in each image. Adds annotations to the database as they are found.

    Args:
        gid_list (list): list of image ids to run detection on
        species (str): string text of the species to identify

    Returns:
        aids_list (list): list of lists of annotation ids detected in each
            image

    CommandLine:
        python -m ibeis.web.apis_detect --test-detect_random_forest --show

    RESTful:
        Method: PUT, GET
        URL:    /api/detect/randomforest/

    Example:
        >>> # DISABLE_DOCTEST
        >>> from ibeis.web.apis_detect import *  # NOQA
        >>> import ibeis
        >>> ibs = ibeis.opendb('testdb1')
        >>> gid_list = ibs.get_valid_gids()[0:2]
        >>> species = ibeis.const.TEST_SPECIES.ZEB_PLAIN
        >>> aids_list = ibs.detect_random_forest(gid_list, species)
        >>> # Visualize results
        >>> if ut.show_was_requested():
        >>>     import plottool_ibeis as pt
        >>>     from ibeis.viz import viz_image
        >>>     for fnum, gid in enumerate(gid_list):
        >>>         viz_image.show_image(ibs, gid, fnum=fnum)
        >>>     pt.show_if_requested()
        >>> # Remove newly detected annotations
        >>> ibs.delete_annots(ut.flatten(aids_list))
    """
    # TODO: Return confidence here as well
    depc = ibs.depc_image
    config = {
        'algo'                   : 'rf',
        'species'                : species,
        'sensitivity'            : 0.2,
        # 'classifier_sensitivity' : 0.64,
        # 'localizer_grid'         : False,
        # 'localizer_sensitivity'  : 0.16,
        # 'labeler_sensitivity'    : 0.42,
        # 'detector_sensitivity'   : 0.08,
    }
    if USE_LOCALIZATIONS:
        results_list = depc.get_property('localizations', gid_list, None, config=config)
        if commit:
            aids_list = ibs.commit_localization_results(gid_list, results_list, note='pyrfdetect')
            return aids_list
    else:
        results_list = depc.get_property('detections', gid_list, None, config=config)
        if commit:
            aids_list = ibs.commit_detection_results(gid_list, results_list, note='pyrfdetect')
            return aids_list


@register_route('/test/review/detect/cnn/yolo/', methods=['GET'])
def review_detection_test(image_uuid=None, result_list=None, callback_url=None,
                          callback_method='POST', **kwargs):
    ibs = current_app.ibs
    if image_uuid is None or result_list is None:
        results_dict = ibs.detection_yolo_test()
        image_uuid = results_dict['image_uuid_list'][0]
        result_list = results_dict['results_list'][0]
    if callback_url is None:
        callback_url = request.args.get('callback_url', url_for('process_detection_html'))
    if callback_method is None:
        callback_method = request.args.get('callback_method', 'POST')
    template_html = review_detection_html(ibs, image_uuid, result_list, callback_url, callback_method, include_jquery=True)
    template_html = '''
        <script src="http://code.jquery.com/jquery-2.2.1.min.js" ia-dependency="javascript"></script>
        %s
    ''' % (template_html, )
    return template_html


@register_ibs_method
@register_api('/test/detect/cnn/yolo/', methods=['GET'])
def detection_yolo_test(ibs):
    from random import shuffle  # NOQA
    gid_list = ibs.get_valid_gids()
    shuffle(gid_list)
    gid_list = gid_list[:3]
    results_dict = ibs.detect_cnn_yolo_json(gid_list)
    return results_dict


@register_ibs_method
@register_api('/test/detect/cnn/lightnet/', methods=['GET'])
def detection_lightnet_test(ibs, config={}):
    from random import shuffle  # NOQA
    gid_list = ibs.get_valid_gids()
    shuffle(gid_list)
    gid_list = gid_list[:3]
    results_dict = ibs.detect_cnn_lightnet_json(gid_list)
    return results_dict


@register_api('/api/review/detect/cnn/yolo/', methods=['GET'])
def review_detection_html(ibs, image_uuid, result_list, callback_url, callback_method='POST', include_jquery=False, config=None):
    """
    Return the detection review interface for a particular image UUID and a list of results for that image.

    Args:
        image_uuid (UUID): the UUID of the image you want to review detections for
        result_list (list of dict): list of detection results returned by the detector
        callback_url (str): URL that the review form will submit to (action) when
            the user is complete with their review
        callback_method (str): HTTP method the review form will submit to (method).
            Defaults to 'POST'

    Returns:
        template (html): json response with the detection web interface in html

    RESTful:
        Method: GET
        URL:    /api/review/detect/cnn/yolo/
    """
    ibs.web_check_uuids(image_uuid_list=[image_uuid])
    gid = ibs.get_image_gids_from_uuid(image_uuid)

    if gid is None:
        return 'INVALID IMAGE UUID'

    default_config = {
        'autointerest'            : False,
        'interest_bypass'         : False,
        'metadata'                : True,
        'metadata_viewpoint'      : False,
        'metadata_quality'        : False,
        'metadata_flags'          : True,
        'metadata_flags_aoi'      : True,
        'metadata_flags_multiple' : False,
        'metadata_species'        : True,
        'metadata_label'          : True,
        'metadata_quickhelp'      : True,
        'parts'                   : False,
        'modes_rectangle'         : True,
        'modes_diagonal'          : True,
        'modes_diagonal2'         : True,
        'staged'                  : False,
    }

    if config is not None:
        default_config.update(config)

    gpath = ibs.get_image_thumbpath(gid, ensure_paths=True, draw_annots=False)
    image = ibs.get_images(gid)
    image_src = appf.embed_image_html(image)
    width, height = ibs.get_image_sizes(gid)

    if width <= 0 or width is None or height <= 0 or height is None:
        vals = (image_uuid, width, height, )
        raise IOError('Image %r for review has either no width or no height (w = %s, h = %s)' % vals)

    annotation_list = []
    for result in result_list:
        quality = result.get('quality', None)
        if quality in [-1, None]:
            quality = 0
        elif quality <= 2:
            quality = 1
        elif quality > 2:
            quality = 2

        viewpoint1 = result.get('viewpoint1', None)
        viewpoint2 = result.get('viewpoint2', None)
        viewpoint3 = result.get('viewpoint3', None)

        if viewpoint1 is None and viewpoint2 is None and viewpoint3 is None:
            viewpoint = result.get('viewpoint', None)
            viewpoint1, viewpoint2, viewpoint3 = appf.convert_viewpoint_to_tuple(viewpoint)

        annotation_list.append({
            'id'      : result.get('id', None),
            'left'    : 100.0 * (result.get('left', result['xtl']) / width),
            'top'     : 100.0 * (result.get('top', result['ytl']) / height),
            'width'   : 100.0 * (result['width'] / width),
            'height'  : 100.0 * (result['height'] / height),
            'species' : result.get('species', result['class']),
            'theta'   : result.get('theta', 0.0),
            'viewpoint1' : viewpoint1,
            'viewpoint2' : viewpoint2,
            'viewpoint3' : viewpoint3,
            'quality'    : quality,
            'multiple'   : 'true' if result.get('multiple', None) == 1 else 'false',
            'interest'   : 'true' if result.get('interest', None) == 1 else 'false',
        })

    species = KEY_DEFAULTS[SPECIES_KEY]

    root_path = dirname(abspath(__file__))
    css_file_list = [
        ['include', 'jquery-ui', 'jquery-ui.min.css'],
        ['include', 'jquery.ui.rotatable', 'jquery.ui.rotatable.css'],
        ['css', 'style.css'],
    ]
    json_file_list = [
        ['include', 'jquery-ui', 'jquery-ui.min.js'],
        ['include', 'jquery.ui.rotatable', 'jquery.ui.rotatable.min.js'],
        ['include', 'bbox_annotator_percent.js'],
        ['javascript', 'script.js'],
        ['javascript', 'turk-detection.js'],
    ]

    if include_jquery:
        json_file_list = [
            ['javascript', 'jquery.min.js'],
        ] + json_file_list

    EMBEDDED_CSS = ''
    EMBEDDED_JAVASCRIPT = ''

    css_template_fmtstr = '<style type="text/css" ia-dependency="css">%s</style>\n'
    json_template_fmtstr = '<script type="text/javascript" ia-dependency="javascript">%s</script>\n'
    for css_file in css_file_list:
        css_filepath_list = [root_path, 'static'] + css_file
        with open(join(*css_filepath_list)) as css_file:
            EMBEDDED_CSS += css_template_fmtstr % (css_file.read(), )

    for json_file in json_file_list:
        json_filepath_list = [root_path, 'static'] + json_file
        with open(join(*json_filepath_list)) as json_file:
            EMBEDDED_JAVASCRIPT += json_template_fmtstr % (json_file.read(), )

    species_rowids = ibs._get_all_species_rowids()
    species_nice_list = ibs.get_species_nice(species_rowids)

    combined_list = sorted(zip(species_nice_list, species_rowids))
    species_nice_list = [ combined[0] for combined in combined_list ]
    species_rowids = [ combined[1] for combined in combined_list ]

    species_text_list = ibs.get_species_texts(species_rowids)
    species_list = list(zip(species_nice_list, species_text_list))
    species_list = [ ('Unspecified', const.UNKNOWN) ] + species_list

    # Collect mapping of species to parts
    aid_list = ibs.get_valid_aids()
    part_species_rowid_list = ibs.get_annot_species_rowids(aid_list)
    part_species_text_list = ibs.get_species_texts(part_species_rowid_list)
    part_rowids_list = ibs.get_annot_part_rowids(aid_list)
    part_types_list = map(ibs.get_part_types, part_rowids_list)

    zipped = list(zip(part_species_text_list, part_types_list))
    species_part_dict = {
        const.UNKNOWN: set([])
    }
    for part_species_text, part_type_list in zipped:
        if part_species_text not in species_part_dict:
            species_part_dict[part_species_text] = set([const.UNKNOWN])
        for part_type in part_type_list:
            species_part_dict[part_species_text].add(part_type)
            species_part_dict[const.UNKNOWN].add(part_type)
    # Add any images that did not get added because they aren't assigned any annotations
    for species_text in species_text_list:
        if species_text not in species_part_dict:
            species_part_dict[species_text] = set([const.UNKNOWN])
    for key in species_part_dict:
        species_part_dict[key] = sorted(list(species_part_dict[key]))
    species_part_dict_json = json.dumps(species_part_dict)

    orientation_flag = '0'
    if species is not None and 'zebra' in species:
        orientation_flag = '1'

    settings_key_list = [
        ('ia-detection-setting-orientation', orientation_flag),
        ('ia-detection-setting-parts-assignments', '1'),
        ('ia-detection-setting-toggle-annotations', '1'),
        ('ia-detection-setting-toggle-parts', '0'),
        ('ia-detection-setting-parts-show', '0'),
        ('ia-detection-setting-parts-hide', '0'),
    ]

    settings = {
        settings_key: request.cookies.get(settings_key, settings_default) == '1'
        for (settings_key, settings_default) in settings_key_list
    }

    return appf.template('turk', 'detection_insert',
                         gid=gid,
                         refer_aid=None,
                         species=species,
                         image_path=gpath,
                         image_src=image_src,
                         config=default_config,
                         settings=settings,
                         annotation_list=annotation_list,
                         species_list=species_list,
                         species_part_dict_json=species_part_dict_json,
                         callback_url=callback_url,
                         callback_method=callback_method,
                         EMBEDDED_CSS=EMBEDDED_CSS,
                         EMBEDDED_JAVASCRIPT=EMBEDDED_JAVASCRIPT)


@register_api('/api/review/detect/cnn/yolo/', methods=['POST'])
def process_detection_html(ibs, **kwargs):
    """
    Process the return from the detection review interface.  Pass the POST result from the detection review form directly to this function unmodified.

    Returns:
        detection results (dict): Same format as `func:start_detect_image`

    RESTful:
        Method: POST
        URL:    /api/review/detect/cnn/yolo/
    """
    gid = int(request.form['detection-gid'])
    image_uuid = ibs.get_image_uuids(gid)
    width, height = ibs.get_image_sizes(gid)
    # Get aids
    annotation_list = json.loads(request.form['ia-detection-data'])

    viewpoint1_list = [
        int(annot['metadata'].get('viewpoint1', -1))
        for annot in annotation_list
    ]
    viewpoint2_list = [
        int(annot['metadata'].get('viewpoint2', -1))
        for annot in annotation_list
    ]
    viewpoint3_list = [
        int(annot['metadata'].get('viewpoint3', -1))
        for annot in annotation_list
    ]
    zipped = list(zip(viewpoint1_list, viewpoint2_list, viewpoint3_list))
    viewpoint_list = [ appf.convert_tuple_to_viewpoint(tup) for tup in zipped ]

    result_list = [
        {
            'id'         : annot['label'],
            'xtl'        : int( width  * (annot['percent']['left']   / 100.0) ),
            'ytl'        : int( height * (annot['percent']['top']    / 100.0) ),
            'left'       : int( width  * (annot['percent']['left']   / 100.0) ),
            'top'        : int( height * (annot['percent']['top']    / 100.0) ),
            'width'      : int( width  * (annot['percent']['width']  / 100.0) ),
            'height'     : int( height * (annot['percent']['height'] / 100.0) ),
            'theta'      : float(annot['angles']['theta']),
            'confidence' : 1.0,
            'class'      : annot['label'],
            'species'    : annot['label'],
            'viewpoint'  : viewpoint,
            'quality'    : annot['metadata']['quality'],
            'multiple'   : annot['metadata']['multiple'],
            'interest'   : annot['highlighted'],

        }
        for annot, viewpoint in list(zip(annotation_list, viewpoint_list))
    ]
    result_dict = {
        'image_uuid_list' : [image_uuid],
        'results_list'    : [result_list],
        'score_list'      : [1.0],
    }
    return result_dict


@register_ibs_method
@accessor_decors.default_decorator
@accessor_decors.getter_1to1
def detect_cnn_json(ibs, gid_list, detect_func, config={}, **kwargs):
    """
    Run animal detection in each image and returns json-ready formatted results, does not return annotations.

    Args:
        gid_list (list): list of image ids to run detection on

    Returns:
        results_dict (list): dict of detection results (not annotations)

    CommandLine:
        python -m ibeis.web.apis_detect --test-detect_cnn_yolo_json

    Example:
        >>> # DISABLE_DOCTEST
        >>> from ibeis.web.apis_detect import *  # NOQA
        >>> import ibeis
        >>> ibs = ibeis.opendb('testdb1')
        >>> gid_list = ibs.get_valid_gids()[0:2]
        >>> results_dict = ibs.detect_cnn_yolo_json(gid_list)
        >>> print(results_dict)
    """
    # TODO: Return confidence here as well
    image_uuid_list = ibs.get_image_uuids(gid_list)
    ibs.assert_valid_gids(gid_list)
    # Get detections from depc
    aids_list = detect_func(gid_list, **config)
    results_list = [
        [
            {
                'id'         : aid,
                'uuid'       : ibs.get_annot_uuids(aid),
                'xtl'        : ibs.get_annot_bboxes(aid)[0],
                'ytl'        : ibs.get_annot_bboxes(aid)[1],
                'left'       : ibs.get_annot_bboxes(aid)[0],
                'top'        : ibs.get_annot_bboxes(aid)[1],
                'width'      : ibs.get_annot_bboxes(aid)[2],
                'height'     : ibs.get_annot_bboxes(aid)[3],
                'theta'      : round(ibs.get_annot_thetas(aid), 4),
                'confidence' : round(ibs.get_annot_detect_confidence(aid), 4),
                'class'      : ibs.get_annot_species_texts(aid),
                'species'    : ibs.get_annot_species_texts(aid),
                'viewpoint'  : ibs.get_annot_viewpoints(aid),
                'quality'    : ibs.get_annot_qualities(aid),
                'multiple'   : ibs.get_annot_multiple(aid),
                'interest'   : ibs.get_annot_interest(aid),
            }
            for aid in aid_list
        ]
        for aid_list in aids_list
    ]
    score_list = [0.0] * len(gid_list)
    # Wrap up results with other information
    results_dict = {
        'image_uuid_list' : image_uuid_list,
        'results_list'    : results_list,
        'score_list'      : score_list,
    }
    return results_dict


@register_ibs_method
@accessor_decors.default_decorator
def detect_cnn_json_wrapper(ibs, image_uuid_list, detect_func, **kwargs):
    """
    Detect with CNN (general).

    REST:
        Method: GET
        URL: /api/detect/cnn/yolo/json/

    Args:
        image_uuid_list (list) : list of image uuids to detect on.
    """
    from ibeis.web.apis_engine import ensure_uuid_list

    # Check UUIDs
    ibs.web_check_uuids(image_uuid_list=image_uuid_list)
    image_uuid_list = ensure_uuid_list(image_uuid_list)
    gid_list = ibs.get_image_gids_from_uuid(image_uuid_list)
    return detect_func(gid_list, **kwargs)


@register_ibs_method
@accessor_decors.default_decorator
@register_api('/api/detect/cnn/yolo/json/', methods=['POST'])
def detect_cnn_yolo_json_wrapper(ibs, image_uuid_list, **kwargs):
    return detect_cnn_json_wrapper(ibs, image_uuid_list, ibs.detect_cnn_yolo_json,
                                   **kwargs)


@register_ibs_method
@accessor_decors.default_decorator
@accessor_decors.getter_1to1
def detect_cnn_yolo_json(ibs, gid_list, config={}, **kwargs):
    return detect_cnn_json(ibs, gid_list, ibs.detect_cnn_yolo,
                           config=config, **kwargs)


@register_ibs_method
@accessor_decors.default_decorator
@accessor_decors.getter_1toM
@register_api('/api/detect/cnn/yolo/', methods=['PUT', 'GET', 'POST'])
def detect_cnn_yolo(ibs, gid_list, commit=True, testing=False, model_tag=None,
                    **kwargs):
    """
    Run animal detection in each image. Adds annotations to the database as they are found.

    Args:
        gid_list (list): list of image ids to run detection on

    Returns:
        aids_list (list): list of lists of annotation ids detected in each
            image

    CommandLine:
        python -m ibeis.web.apis_detect --test-detect_cnn_yolo --show

    RESTful:
        Method: PUT, GET
        URL:    /api/detect/cnn/yolo/

    Example:
        >>> # DISABLE_DOCTEST
        >>> from ibeis.web.apis_detect import *  # NOQA
        >>> import ibeis
        >>> ibs = ibeis.opendb('PZ_MTEST')
        >>> gid_list = ibs.get_valid_gids()[:5]
        >>> aids_list = ibs.detect_cnn_yolo(gid_list)
        >>> if ut.show_was_requested():
        >>>     import plottool_ibeis as pt
        >>>     from ibeis.viz import viz_image
        >>>     for fnum, gid in enumerate(gid_list):
        >>>         viz_image.show_image(ibs, gid, fnum=fnum)
        >>>     pt.show_if_requested()
        >>> # Remove newly detected annotations
        >>> ibs.delete_annots(ut.flatten(aids_list))
    """
    # TODO: Return confidence here as well
    depc = ibs.depc_image
    config = {
        'algo'                   : 'yolo',
        'sensitivity'            : 0.2,
        # 'classifier_sensitivity' : 0.64,
        # 'localizer_grid'         : False,
        # 'localizer_sensitivity'  : 0.16,
        # 'labeler_sensitivity'    : 0.42,
        # 'detector_sensitivity'   : 0.08,
    }
    if model_tag is not None:
        config['config_filepath'] = model_tag
        config['weight_filepath'] = model_tag
    config_str_list = ['config_filepath', 'weight_filepath'] + list(config.keys())
    for config_str in config_str_list:
        if config_str in kwargs:
            config[config_str] = kwargs[config_str]
    if USE_LOCALIZATIONS:
        if testing:
            depc.delete_property('localizations', gid_list, config=config)
        results_list = depc.get_property('localizations', gid_list, None, config=config)
        if commit:
            # labeler_config = config.copy()
            # labeler_config['labeler_weight_filepath'] = 'candidacy'
            # viewpoints_list = depc.get_property('localizations_labeler', gid_list, 'viewpoint', config=labeler_config)
            viewpoints_list = None
            aids_list = ibs.commit_localization_results(gid_list, results_list, viewpoints_list=viewpoints_list, note='cnnyolodetect')
            return aids_list
    else:
        if testing:
            depc.delete_property('detections', gid_list, config=config)
        results_list = depc.get_property('detections', gid_list, None, config=config)
        if commit:
            aids_list = ibs.commit_detection_results(gid_list, results_list, note='cnnyolodetect')
            return aids_list


@register_ibs_method
@accessor_decors.default_decorator
@accessor_decors.getter_1to1
@register_api('/api/detect/cnn/yolo/exists/', methods=['GET'], __api_plural_check__=False)
def detect_cnn_yolo_exists(ibs, gid_list, testing=False):
    """
    Check to see if a detection has been completed.

    Args:
        gid_list (list): list of image ids to run detection on

    Returns:
        flag_list (list): list of flags for if the detection has been run on
            the image

    CommandLine:
        python -m ibeis.web.apis_detect --test-detect_cnn_yolo_exists

    RESTful:
        Method: GET
        URL:    /api/detect/cnn/yolo/exists/

    Example:
        >>> # DISABLE_DOCTEST
        >>> from ibeis.web.apis_detect import *  # NOQA
        >>> import ibeis
        >>> ibs = ibeis.opendb('PZ_MTEST')
        >>> gid_list = ibs.get_valid_gids()
        >>> depc = ibs.depc_image
        >>> aids_list = ibs.detect_cnn_yolo(gid_list[:3], testing=True)
        >>> result = ibs.detect_cnn_yolo_exists(gid_list[:5])
        >>> ibs.delete_annots(ut.flatten(aids_list))
        >>> print(result)
        [True, True, True, False, False]
    """
    depc = ibs.depc_image
    config = {
        'algo'                   : 'yolo',
        'sensitivity'            : 0.2,
        # 'classifier_sensitivity' : 0.64,
        # 'localizer_grid'         : False,
        # 'localizer_sensitivity'  : 0.16,
        # 'labeler_sensitivity'    : 0.42,
        # 'detector_sensitivity'   : 0.08,
    }
    if USE_LOCALIZATIONS:
        score_list = depc.get_property('localizations', gid_list, 'score', ensure=False, config=config)
    else:
        score_list = depc.get_property('detections', gid_list, 'score', ensure=False, config=config)
    flag_list = [ score is not None for score in score_list ]
    return flag_list


@register_ibs_method
@accessor_decors.default_decorator
@register_api('/api/detect/cnn/lightnet/json/', methods=['POST'])
def detect_cnn_lightnet_json_wrapper(ibs, image_uuid_list, **kwargs):
    return detect_cnn_json_wrapper(ibs, image_uuid_list, ibs.detect_cnn_lightnet_json,
                                   **kwargs)


@register_ibs_method
@accessor_decors.default_decorator
@accessor_decors.getter_1to1
def detect_cnn_lightnet_json(ibs, gid_list, config={}, **kwargs):
    return detect_cnn_json(ibs, gid_list, ibs.detect_cnn_lightnet,
                           config=config, **kwargs)


@register_ibs_method
@accessor_decors.default_decorator
@accessor_decors.getter_1toM
@register_api('/api/detect/cnn/lightnet/', methods=['PUT', 'GET', 'POST'])
def detect_cnn_lightnet(ibs, gid_list, commit=True, testing=False, model_tag=None,
                        **kwargs):
    """
    Run animal detection in each image. Adds annotations to the database as they are found.

    Args:
        gid_list (list): list of image ids to run detection on

    Returns:
        aids_list (list): list of lists of annotation ids detected in each
            image

    CommandLine:
        python -m ibeis.web.apis_detect --test-detect_cnn_lightnet --show

    RESTful:
        Method: PUT, GET
        URL:    /api/detect/cnn/lightnet/

    Example:
        >>> # DISABLE_DOCTEST
        >>> from ibeis.web.apis_detect import *  # NOQA
        >>> import ibeis
        >>> ibs = ibeis.opendb('PZ_MTEST')
        >>> gid_list = ibs.get_valid_gids()[:5]
        >>> aids_list = ibs.detect_cnn_lightnet(gid_list)
        >>> if ut.show_was_requested():
        >>>     import plottool_ibeis as pt
        >>>     from ibeis.viz import viz_image
        >>>     for fnum, gid in enumerate(gid_list):
        >>>         viz_image.show_image(ibs, gid, fnum=fnum)
        >>>     pt.show_if_requested()
        >>> # Remove newly detected annotations
        >>> ibs.delete_annots(ut.flatten(aids_list))
    """
    # TODO: Return confidence here as well
    depc = ibs.depc_image
    config = {
        'algo'        : 'lightnet',
        'sensitivity' : 0.75,
        'nms'         : True,
        'nms_thresh'  : 0.4,
    }
    if model_tag is not None:
        config['weight_filepath'] = model_tag
    config_str_list = ['weight_filepath'] + list(config.keys())
    for config_str in config_str_list:
        if config_str in kwargs:
            config[config_str] = kwargs[config_str]
    if testing:
        depc.delete_property('localizations', gid_list, config=config)
    results_list = depc.get_property('localizations', gid_list, None, config=config)
    if commit:
        labeler_config = config.copy()
        labeler_config['labeler_weight_filepath'] = 'candidacy'
        viewpoints_list = depc.get_property('localizations_labeler', gid_list, 'viewpoint', config=labeler_config)
        aids_list = ibs.commit_localization_results(gid_list, results_list, viewpoints_list=viewpoints_list, note='cnnlightnetdetect')
        return aids_list


@register_ibs_method
def commit_localization_results(ibs, gid_list, results_list, viewpoints_list=None, note=None,
                                update_json_log=True):
    if viewpoints_list is None:
        viewpoints_list = [None] * len(gid_list)

    zipped_list = list(zip(gid_list, results_list, viewpoints_list))
    aids_list = []
    for gid, (score, bbox_list, theta_list, conf_list, class_list), viewpoint_list in zipped_list:
        if viewpoint_list is not None:
            assert len(viewpoint_list) == len(bbox_list)

        num = len(bbox_list)
        notes_list = None if note is None else [note] * num
        aid_list = ibs.add_annots(
            [gid] * num,
            bbox_list,
            theta_list,
            class_list,
            viewpoint_list=viewpoint_list,
            detect_confidence_list=conf_list,
            notes_list=notes_list,
            quiet_delete_thumbs=True,
            skip_cleaning=True
        )
        aids_list.append(aid_list)
    ibs._clean_species()
    if update_json_log:
        aid_list = ut.flatten(aids_list)
        ibs.log_detections(aid_list)
    return aids_list


@register_ibs_method
def commit_detection_results(ibs, gid_list, results_list, note=None,
                             update_json_log=True):
    zipped_list = list(zip(gid_list, results_list))
    aids_list = []
    for gid, (score, bbox_list, theta_list, species_list, viewpoint_list, conf_list) in zipped_list:
        num = len(bbox_list)
        notes_list = None if note is None else [note] * num
        aid_list = ibs.add_annots(
            [gid] * num,
            bbox_list,
            theta_list,
            species_list,
            detect_confidence_list=conf_list,
            notes_list=notes_list,
            quiet_delete_thumbs=True,
            skip_cleaning=True
        )
        ibs.set_annot_viewpoints(aid_list, viewpoint_list)
        # TODO ibs.set_annot_viewpoint_code(aid_list, viewpoint_list)
        aids_list.append(aid_list)
    ibs._clean_species()
    if update_json_log:
        aid_list = ut.flatten(aids_list)
        ibs.log_detections(aid_list)
    return aids_list


@register_ibs_method
def commit_detection_results_filtered(ibs, gid_list, filter_species_list=None,
                                      filter_viewpoint_list=None, note=None,
                                      update_json_log=True):
    depc = ibs.depc_image
    results_list = depc.get_property('detections', gid_list, None)
    zipped_list = list(zip(gid_list, results_list))
    aids_list = []
    for gid, (score, bbox_list, theta_list, species_list, viewpoint_list, conf_list) in zipped_list:
        aid_list = []
        result_list = list(zip(bbox_list, theta_list, species_list, viewpoint_list, conf_list))
        for bbox, theta, species, viewpoint, conf in result_list:
            if not (filter_species_list is None or species in filter_species_list):
                continue
            if not (filter_viewpoint_list is None or viewpoint in filter_viewpoint_list):
                continue
            note_ = None if note is None else [note]
            temp_list = ibs.add_annots(
                [gid],
                [bbox],
                [theta],
                [species],
                detect_confidence_list=[conf],
                notes_list=note_,
                quiet_delete_thumbs=True,
                skip_cleaning=True
            )
            aid = temp_list[0]
            ibs.set_annot_viewpoints([aid], [viewpoint])
            # TODO ibs.set_annot_viewpoint_code([aid], [viewpoint])
            aid_list.append(aid)
        aids_list.append(aid_list)
    ibs._clean_species()
    if update_json_log:
        aid_list = ut.flatten(aids_list)
        ibs.log_detections(aid_list)
    return aids_list


@register_ibs_method
def log_detections(ibs, aid_list, fallback=True):
    import time
    import os
    json_log_path = ibs.get_logdir_local()
    json_log_filename = 'detections.json'
    json_log_filepath = os.path.join(json_log_path, json_log_filename)
    print('Logging detections added to: %r' % (json_log_filepath, ))

    try:
        # Log has never been made, create one
        if not os.path.exists(json_log_filepath):
            json_dict = {
                'updates': [],
            }
            json_str = ut.to_json(json_dict, pretty=True)
            with open(json_log_filepath, 'w') as json_log_file:
                json_log_file.write(json_str)
        # Get current log state
        with open(json_log_filepath, 'r') as json_log_file:
            json_str = json_log_file.read()
        json_dict = ut.from_json(json_str)
        # Get values
        db_name = ibs.get_db_name()
        db_init_uuid = ibs.get_db_init_uuid()
        # Zip all the updates together and write to updates list in dictionary
        gid_list = ibs.get_annot_gids(aid_list)
        bbox_list = ibs.get_annot_bboxes(aid_list)
        theta_list = ibs.get_annot_thetas(aid_list)
        zipped = list(zip(aid_list, gid_list, bbox_list, theta_list))
        for aid, gid, bbox, theta in zipped:
            json_dict['updates'].append({
                'time_unixtime': time.time(),
                'db_name': db_name,
                'db_init_uuid': db_init_uuid,
                'image_rowid': gid,
                'annot_rowid': aid,
                'annot_bbox': bbox,
                'annot_theta': theta,
            })
        # Write new log state
        json_str = ut.to_json(json_dict, pretty=True)
        with open(json_log_filepath, 'w') as json_log_file:
            json_log_file.write(json_str)
    except:
        if fallback:
            print('WRITE DETECTION.JSON FAILED - ATTEMPTING FALLBACK')
            ut.delete(json_log_filepath)
            ibs.log_detections(aid_list, fallback=False)
        else:
            print('WRITE DETECTION.JSON FAILED - FALLBACK FAILED')


@register_ibs_method
@accessor_decors.default_decorator
@register_api('/api/detect/species/enabled/', methods=['GET'], __api_plural_check__=False)
def has_species_detector(ibs, species_text):
    """
    TODO: extend to use non-constant species.

    RESTful:
        Method: GET
        URL:    /api/detect/species/enabled/
    """
    # FIXME: infer this
    return species_text in const.SPECIES_WITH_DETECTORS


@register_ibs_method
@accessor_decors.default_decorator
@register_api('/api/detect/species/', methods=['GET'], __api_plural_check__=False)
def get_species_with_detectors(ibs):
    """
    Get valid species for detection.

    RESTful:
        Method: GET
        URL:    /api/detect/species/
    """
    # FIXME: infer this
    return const.SPECIES_WITH_DETECTORS


@register_ibs_method
@accessor_decors.default_decorator
@register_api('/api/detect/species/working/', methods=['GET'], __api_plural_check__=False)
def get_working_species(ibs):
    """
    Get working species for detection.

    RESTful:
        Method: GET
        URL:    /api/detect/species/working/
    """
    RESTRICT_TO_ONLY_SPECIES_WITH_DETECTORS = ut.get_argflag('--no-allspecies')

    species_nice_list = ibs.get_all_species_nice()
    species_text_list = ibs.get_all_species_texts()
    species_tup_list = list(zip(species_nice_list, species_text_list))
    if RESTRICT_TO_ONLY_SPECIES_WITH_DETECTORS:
        working_species_tups = [
            species_tup
            for species_tup in species_tup_list
            if ibs.has_species_detector(species_tup[1])
        ]
    else:
        working_species_tups = species_tup_list
    return working_species_tups


@register_ibs_method
@accessor_decors.default_decorator
@register_api('/api/detect/whaleSharkInjury/', methods=['PUT', 'GET'])
def detect_ws_injury(ibs, gid_list):
    """
    Classify if a whale shark is injured.

    Args:
        gid_list (list): list of image ids to run classification on

    Returns:
        result_list (dictionary): predictions is list of strings representing a possible tag.
            confidences is a list of floats of correspoinding cofidence to the prediction

    """
    from ibeis.scripts import labelShark
    labels = labelShark.classifyShark(ibs, gid_list)
    return labels

if __name__ == '__main__':
    """
    CommandLine:
        python -m ibeis.web.app
        python -m ibeis.web.app --allexamples
        python -m ibeis.web.app --allexamples --noface --nosrc
    """
    import multiprocessing
    multiprocessing.freeze_support()  # for win32
    import utool as ut  # NOQA
    ut.doctest_funcs()



""" Implements ProcessActor """
from concurrent.futures import _base
from concurrent.futures import process
from multiprocessing.connection import wait
from ibeis.web.futures_utils import _base_actor
import os
import queue
import weakref
import threading
import multiprocessing


# Most of this code is duplicated from the concurrent.futures.thread and
# concurrent.futures.process modules, writen by Brian Quinlan. The main
# difference is that we expose an `Actor` class which can be inherited from and
# provides the `executor` classmethod. This creates an asynchronously
# maintained instance of this class in a separate thread/process

__author__ = 'Jon Crall (erotemic@gmail.com)'


def _process_actor_eventloop(_call_queue, _result_queue, _ActorClass, *args,
                             **kwargs):
    """
    actor event loop run in a separate process.

    Creates the instance of the actor (passing in the required *args, and
    **kwargs). Then the eventloop starts and feeds the actor messages from the
    _call_queue. Results are placed in the _result_queue, which are then placed
    in Future objects.
    """

    actor = _ActorClass(*args, **kwargs)
    while True:
        call_item = _call_queue.get(block=True)
        if call_item is None:
            # Wake up queue management thread
            _result_queue.put(os.getpid())
            return
        try:
            r = actor.handle(call_item.message)
        except BaseException as e:
            exc = process._ExceptionWithTraceback(e, e.__traceback__)
            _result_queue.put(process._ResultItem(
                call_item.work_id, exception=exc))
        else:
            _result_queue.put(process._ResultItem(
                call_item.work_id, result=r))


class _WorkItem(object):
    def __init__(self, future, message):
        self.future = future
        self.message = message


class _CallItem(object):
    def __init__(self, work_id, message):
        self.work_id = work_id
        self.message = message


def _add_call_item_to_queue(pending_work_items,
                            work_ids,
                            call_queue):
    """Fills call_queue with _WorkItems from pending_work_items.

    This function never blocks.

    Args:
        pending_work_items: A dict mapping work ids to _WorkItems e.g.
            {5: <_WorkItem...>, 6: <_WorkItem...>, ...}
        work_ids: A queue.Queue of work ids e.g. Queue([5, 6, ...]). Work ids
            are consumed and the corresponding _WorkItems from
            pending_work_items are transformed into _CallItems and put in
            call_queue.
        call_queue: A multiprocessing.Queue that will be filled with _CallItems
            derived from _WorkItems.
    """
    while True:
        if call_queue.full():
            return
        try:
            work_id = work_ids.get(block=False)
        except queue.Empty:
            return
        else:
            work_item = pending_work_items[work_id]

            if work_item.future.set_running_or_notify_cancel():
                call_queue.put(_CallItem(work_id,
                                         work_item.message),
                               block=True)
            else:
                del pending_work_items[work_id]
                continue


def _queue_management_worker(executor_reference,
                             _manager,
                             pending_work_items,
                             work_ids_queue,
                             _call_queue,
                             _result_queue):
    """Manages the communication between this process and the worker processes."""
    executor = None

    def shutting_down():
        return process._shutdown or executor is None or executor._shutdown_thread

    def shutdown_worker():
        # This is an upper bound
        if _manager.is_alive():
            _call_queue.put_nowait(None)
        # Release the queue's resources as soon as possible.
        _call_queue.close()
        # If .join() is not called on the created processes then
        # some multiprocessing.Queue methods may deadlock on Mac OS X.
        _manager.join()

    reader = _result_queue._reader

    while True:
        _add_call_item_to_queue(pending_work_items,
                                work_ids_queue,
                                _call_queue)

        sentinel = _manager.sentinel
        assert sentinel
        ready = wait([reader, sentinel])
        if reader in ready:
            result_item = reader.recv()
        else:
            # Mark the process pool broken so that submits fail right now.
            executor = executor_reference()
            if executor is not None:
                executor._broken = True
                executor._shutdown_thread = True
                executor = None
            # All futures in flight must be marked failed
            for work_id, work_item in pending_work_items.items():
                work_item.future.set_exception(
                    process.BrokenProcessPool(
                        "A process in the process pool was "
                        "terminated abruptly while the future was "
                        "running or pending."
                    ))
                # Delete references to object. See issue16284
                del work_item
            pending_work_items.clear()
            # Terminate remaining workers forcibly: the queues or their
            # locks may be in a dirty state and block forever.
            _manager.terminate()
            shutdown_worker()
            return
        if isinstance(result_item, int):
            # Clean shutdown of a worker using its PID
            # (avoids marking the executor broken)
            assert shutting_down()
            _manager.join()
            if _manager is None:
                shutdown_worker()
                return
        elif result_item is not None:
            work_item = pending_work_items.pop(result_item.work_id, None)
            # work_item can be None if another process terminated (see above)
            if work_item is not None:
                if result_item.exception:
                    work_item.future.set_exception(result_item.exception)
                else:
                    work_item.future.set_result(result_item.result)
                # Delete references to object. See issue16284
                del work_item
        # Check whether we should start shutting down.
        executor = executor_reference()
        # No more work items can be added if:
        #   - The interpreter is shutting down OR
        #   - The executor that owns this worker has been collected OR
        #   - The executor that owns this worker has been shutdown.
        if shutting_down():
            try:
                # Since no new work items can be added, it is safe to shutdown
                # this thread if there are no pending work items.
                if not pending_work_items:
                    shutdown_worker()
                    return
            except queue.Full:
                # This is not a problem: we will eventually be woken up (in
                # _result_queue.get()) and be able to send a sentinel again.
                pass
        executor = None


class ProcessActorExecutor(_base_actor.ActorExecutor):

    def __init__(self, _ActorClass, *args, **kwargs):
        process._check_system_limits()

        self._ActorClass = _ActorClass
        # todo: If we want to cancel futures we need to give the task_queue a
        # maximum size
        self._call_queue = multiprocessing.JoinableQueue()
        self._call_queue._ignore_epipe = True
        self._result_queue = multiprocessing.Queue()
        self._work_ids = queue.Queue()
        self._queue_management_thread = None

        # We only maintain one process for our actor
        self._manager = None

        # Shutdown is a two-step process.
        self._shutdown_thread = False
        self._shutdown_lock = threading.Lock()
        self._broken = False
        self._queue_count = 0
        self._pending_work_items = {}

        self._did_initialize = False

        if args or kwargs:
            # If given actor initialization args we must start the Actor
            # immediately. Otherwise just wait until we get a message
            print('Init with args')
            print('args = %r' % (args,))
            self._initialize_actor(*args, **kwargs)

    def post(self, message):
        with self._shutdown_lock:
            if self._broken:
                raise process.BrokenProcessPool(
                    'A child process terminated '
                    'abruptly, the process pool is not usable anymore')
            if self._shutdown_thread:
                raise RuntimeError('cannot schedule new futures after shutdown')

            f = _base.Future()
            w = _WorkItem(f, message)

            self._pending_work_items[self._queue_count] = w
            self._work_ids.put(self._queue_count)
            self._queue_count += 1
            # Wake up queue management thread
            self._result_queue.put(None)

            self._start_queue_management_thread()
            return f
    post.__doc__ = _base_actor.ActorExecutor.post.__doc__

    def _start_queue_management_thread(self):
        # When the executor gets lost, the weakref callback will wake up
        # the queue management thread.

        def weakref_cb(_, q=self._result_queue):
            q.put(None)

        if self._queue_management_thread is None:
            # Start the processes so that their sentinel are known.
            self._initialize_actor()
            self._queue_management_thread = threading.Thread(
                    target=_queue_management_worker,
                    args=(weakref.ref(self, weakref_cb),
                          self._manager,
                          self._pending_work_items,
                          self._work_ids,
                          self._call_queue,
                          self._result_queue))
            self._queue_management_thread.daemon = True
            self._queue_management_thread.start()
            # use structures already in futures as much as possible
            process._threads_queues[self._queue_management_thread] = self._result_queue

    def _initialize_actor(self, *args, **kwargs):
        if self._manager is None:
            assert self._did_initialize is False, 'only initialize actor once'
            self._did_initialize = True
            # We only maintain one thread process for an actor
            self._manager = multiprocessing.Process(
                    target=_process_actor_eventloop,
                    args=(self._call_queue,
                          self._result_queue, self._ActorClass) + args,
                    kwargs=kwargs)
            self._manager.start()

    def shutdown(self, wait=True):
        with self._shutdown_lock:
            self._shutdown_thread = True
        if self._queue_management_thread:
            # Wake up queue management thread
            self._result_queue.put(None)
            if wait:
                self._queue_management_thread.join()
        # To reduce the risk of opening too many files, remove references to
        # objects that use file descriptors.
        self._queue_management_thread = None
        self._call_queue = None
        self._result_queue = None
        self._manager = None
    shutdown.__doc__ = _base.Executor.shutdown.__doc__


class ProcessActor(_base_actor.Actor):

    @classmethod
    def executor(cls, *args, **kwargs):
        return ProcessActorExecutor(cls, *args, **kwargs)
    # executor.__doc__ = _base_actor.Actor.executor.__doc___

# ProcessActor.__doc__ = _base_actor.Actor.__doc___



# -*- coding: utf-8 -*-
"""
displays results from harness

TODO: save a testres variable so reloading and regenration becomes easier.
"""
from __future__ import absolute_import, division, print_function, unicode_literals
import numpy as np
import six
import utool as ut
from six.moves import map, range
print, rrr, profile = ut.inject2(__name__)


def get_diffranks(rank_mat, qaids):
    """ Find rows which scored differently over the various configs
    FIXME: duplicated
    """
    isdiff_flags = [not np.all(row == row[0]) for row in rank_mat]
    diff_aids    = ut.compress(qaids, isdiff_flags)
    diff_rank    = rank_mat.compress(isdiff_flags, axis=0)
    diff_qxs     = np.where(isdiff_flags)[0]
    return diff_aids, diff_rank, diff_qxs


def get_diffmat_str(rank_mat, qaids, nConfig):
    from itertools import chain
    diff_aids, diff_rank, diff_qxs = get_diffranks(rank_mat, qaids)
    # Find columns that ore strictly better than other columns
    #def find_strictly_better_columns(diff_rank):
    #    colmat = diff_rank.T
    #    pairwise_betterness_ranks = np.array([np.sum(col <= colmat, axis=1) / len(col) for col in colmat], dtype=np.float).T
    diff_mat = np.vstack((diff_aids, diff_rank.T)).T
    col_lbls = list(chain(['qaid'], map(lambda x: 'cfg%d_rank' % x, range(nConfig))))
    col_type  = list(chain([int], [int] * nConfig))
    header = 'diffmat'
    diff_matstr = ut.numpy_to_csv(diff_mat, col_lbls, header, col_type)
    return diff_matstr


def print_latexsum(ibs, testres, verbose=True):
    r"""
    Args:
        ibs (IBEISController):  ibeis controller object
        testres (?):

    CommandLine:
        python -m ibeis.expt.experiment_printres --exec-print_latexsum
        python -m ibeis.scripts.gen_cand_expts --exec-gen_script

        python -m ibeis --tf print_latexsum -t candidacy --db PZ_Master0 -a controlled --rank-lt-list=1,5,10,100
        python -m ibeis --tf print_latexsum -t candidacy --db PZ_MTEST -a controlled --rank-lt-list=1,5,10,100

    Example:
        >>> # SCRIPT
        >>> from ibeis.expt.experiment_printres import *  # NOQA
        >>> from ibeis.init import main_helpers
        >>> ibs, testres = main_helpers.testdata_expts()
        >>> tabular_str2 = print_latexsum(ibs, testres)
    """
    print('==========================')
    print('[harn] LaTeX: %s' % testres.testnameid)
    print('==========================')
    # Create configuration latex table
    X_LIST = testres.get_X_LIST()
    criteria_lbls = [r'#ranks $\leq$ %d' % X for X in X_LIST]
    dbname = ibs.get_dbname()
    cfg_score_title = dbname + ' rank scores'
    nLessX_dict = testres.get_nLessX_dict()
    cfgscores = np.array([nLessX_dict[int(X)] for X in X_LIST]).T

    # For mat row labels
    row_lbls = testres.get_short_cfglbls()
    # Order cdf list by rank0
    row_lbls = ut.sortedby(row_lbls, cfgscores.T[0], reverse=True)
    cfgscores = np.array(ut.sortedby(cfgscores.tolist(), cfgscores.T[0], reverse=True))

    cmdaug = testres.get_title_aug()
    #if testres.common_acfg is not None:
    #    cfgname = testres.common_acfg['common']['_cfgname']
    #    cmdaug += '_' + cfgname
    #if hasattr(testres, 'common_cfgdict'):
    #    cmdaug += '_' + (testres.common_cfgdict['_cfgname'])
    #    cfg_score_title += ' ' + cmdaug

    tabular_kwargs = dict(
        title=cfg_score_title,
        out_of=testres.nQuery,
        bold_best=True,
        flip=False,
        SHORTEN_ROW_LBLS=False
    )
    col_lbls = criteria_lbls
    tabular_str = ut.util_latex.make_score_tabular(
        row_lbls, col_lbls, cfgscores, **tabular_kwargs)
    #latex_formater.render(tabular_str)
    cmdname = ut.latex_sanitize_command_name('Expmt' + ibs.get_dbname() + '_' + cmdaug + 'Table')
    tabular_str2 = ut.latex_newcommand(cmdname, tabular_str)
    print(tabular_str2)
    return tabular_str2


@profile
def print_results(ibs, testres, **kwargs):
    """
    Prints results from an experiment harness run.
    Rows store different qaids (query annotation ids)
    Cols store different configurations (algorithm parameters)

    TODO: join acfgs

    Args:
        ibs (IBEISController):  ibeis controller object
        testres (test_result.TestResult):

    CommandLine:
        python dev.py -e print --db PZ_MTEST \
            -a default:dpername=1,qpername=[1,2]  -t default:fg_on=False
        python dev.py -e print -t best --db seals2 --allgt --vz
        python dev.py -e print --db PZ_MTEST --allgt -t custom \
            --print-confusion-stats
        python dev.py -e print --db PZ_MTEST --allgt --noqcache \
            --index 0:10:2 -t custom:rrvsone_on=True --print-confusion-stats
        python dev.py -e print --db PZ_MTEST --allgt --noqcache --qaid4 \
            -t custom:rrvsone_on=True --print-confusion-stats
        python -m ibeis print_results -t default --db PZ_MTEST -a ctrl
        python -m ibeis print_results -t default --db PZ_MTEST -a ctrl
        python -m ibeis print_results --db PZ_MTEST -a default
            -t default:lnbnn_on=True default:lnbnn_on=False,bar_l2_on=True \
            default:lnbnn_on=False,normonly_on=True

    CommandLine:
        python -m ibeis.expt.experiment_printres --test-print_results
        utprof.py -m ibeis.expt.experiment_printres --test-print_results

    Example:
        >>> # DISABLE_DOCTEST
        >>> from ibeis.expt.experiment_printres import *  # NOQA
        >>> from ibeis.init import main_helpers
        >>> ibs, testres = main_helpers.testdata_expts(
        >>>     'pz_mtest', a='default:dpername=1,qpername=[1,2]',
        >>>     t='default:fg_on=false')
        >>> result = print_results(ibs, testres)
        >>> print(result)
    """

    tup = ut.dict_take(testres.__dict__,
                       ['cfg_list', 'cfgx2_cmsinfo', 'testnameid',
                        'cfgx2_lbl', 'cfgx2_qreq_'])
    (cfg_list, cfgx2_cmsinfo, testnameid, cfgx2_lbl, cfgx2_qreq_) = tup

    # join_acfgs = kwargs.get('join_acfgs', False)

    print(' --- PRINT RESULTS ---')
    # print(' use --rank-lt-list=1,5 to specify X_LIST')
    if True:
        # Num of ranks less than to score
        X_LIST = testres.get_X_LIST()
        #X_LIST = [1, 5]

        #nConfig = len(cfg_list)
        #nQuery = len(testres.qaids)
        cfgx2_nQuery = list(map(len, testres.cfgx2_qaids))
        #cfgx2_qx2_ranks = testres.get_infoprop_list('qx2_gt_rank')
        #--------------------

        # A positive scorediff indicates the groundtruth was better than the
        # groundfalse scores
        #istrue_list  = [scorediff > 0 for scorediff in scorediffs_mat]
        #isfalse_list = [~istrue for istrue in istrue_list]

        #------------
        # Build Colscore
        nLessX_dict = testres.get_nLessX_dict()

        # cfgx2_hist, edges = testres.get_rank_histograms(bins=X_LIST + [np.inf],
        #                                                join_acfgs=join_acfgs)
        # cfgx2_cumsum = cfgx2_hist.cumsum(axis=1)

        #------------
        best_rankscore_summary = []
        #to_intersect_list = []
        # print each configs scores less than X=thresh
        for X, cfgx2_nLessX in six.iteritems(nLessX_dict):
            max_nLessX = cfgx2_nLessX.max()
            bestX_cfgx_list = np.where(cfgx2_nLessX == max_nLessX)[0]
            best_rankscore = '[cfg*] %d cfg(s) scored ' % len(bestX_cfgx_list)
            # FIXME
            best_rankscore += rankscore_str(X, max_nLessX,
                                            cfgx2_nQuery[bestX_cfgx_list[0]])
            best_rankscore_summary += [best_rankscore]

    @ut.argv_flag_dec
    def intersect_hack():
        failed = testres.rank_mat > 0
        colx2_failed = [np.nonzero(failed_col)[0] for failed_col in failed.T]
        #failed_col2_only = np.setdiff1d(colx2_failed[1], colx2_failed[0])
        #failed_col2_only_aids = ut.take(testres.qaids, failed_col2_only)
        failed_col1_only = np.setdiff1d(colx2_failed[0], colx2_failed[1])
        failed_col1_only_aids = ut.take(testres.qaids, failed_col1_only)
        gt_aids1 = ibs.get_annot_groundtruth(failed_col1_only_aids,
                                             daid_list=testres.cfgx2_qreq_[0].daids)
        gt_aids2 = ibs.get_annot_groundtruth(failed_col1_only_aids,
                                             daid_list=testres.cfgx2_qreq_[1].daids)

        qaids_expt = failed_col1_only_aids
        gt_avl_aids1 = ut.flatten(gt_aids1)
        gt_avl_aids2 = list(set(ut.flatten(gt_aids2)).difference(gt_avl_aids1))

        ibs.print_annotconfig_stats(qaids_expt, gt_avl_aids1)
        ibs.print_annotconfig_stats(qaids_expt, gt_avl_aids2)
        #jsontext = ut.to_json({
        #    'qaids': list(qaids_expt),
        #    'dinclude_aids1': list(gt_aids_expt1),
        #    'dinclude_aids2': list(gt_aids_expt2),
        #})
        #annotation_configs.varysize_pzm
        #from ibeis.expt import annotation_configs

        acfg = testres.acfg_list[0]
        import copy
        acfg1 = copy.deepcopy(acfg)
        acfg2 = copy.deepcopy(acfg)
        acfg1['qcfg']['min_pername'] = None
        acfg2['qcfg']['min_pername'] = None
        acfg1['dcfg']['min_pername'] = None
        acfg2['dcfg']['min_gt_per_name'] = None

        acfg1['qcfg']['default_aids'] = qaids_expt
        acfg1['dcfg']['gt_avl_aids'] = gt_avl_aids1
        acfg2['qcfg']['default_aids'] = qaids_expt
        acfg2['dcfg']['gt_avl_aids'] = gt_avl_aids2

        from ibeis.init import filter_annots
        from ibeis.expt import experiment_helpers

        annots1 = filter_annots.expand_acfgs(ibs, acfg1, verbose=True)
        annots2 = filter_annots.expand_acfgs(ibs, acfg2, verbose=True)

        acfg_name_list = dict(  # NOQA
            acfg_list=[acfg1, acfg2],
            expanded_aids_list=[annots1, annots2],
        )
        test_cfg_name_list = ['candidacy_k']
        cfgdict_list, pipecfg_list = experiment_helpers.get_pipecfg_list(
            test_cfg_name_list, ibs=ibs)

        t1, t2 = testres_list  # NOQA
    #intersect_hack()

    #@ut.argv_flag_dec
    #def print_rowlbl():
    #    print('=====================')
    #    print('[harn] Row/Query Labels: %s' % testnameid)
    #    print('=====================')
    #    print('[harn] queries:\n%s' % '\n'.join(qx2_lbl))
    #print_rowlbl()
    #------------

    @ut.argv_flag_dec
    def print_collbl():
        print('=====================')
        print('[harn] Col/Config Labels: %s' % testnameid)
        print('=====================')
        enum_cfgx2_lbl = ['%2d) %s' % (count, cfglbl)
                            for count, cfglbl in enumerate(cfgx2_lbl)]
        print('[harn] cfglbl:\n%s' % '\n'.join(enum_cfgx2_lbl))
    print_collbl()

    #------------

    @ut.argv_flag_dec
    def print_cfgstr():
        print('=====================')
        print('[harn] Config Strings: %s' % testnameid)
        print('=====================')
        cfgstr_list = [query_cfg.get_cfgstr() for query_cfg in cfg_list]
        enum_cfgstr_list = ['%2d) %s' % (count, cfgstr)
                            for count, cfgstr in enumerate(cfgstr_list)]
        print('\n[harn] cfgstr:\n%s' % '\n'.join(enum_cfgstr_list))
    print_cfgstr(**kwargs)

    @ut.argv_flag_dec()
    def print_colscore():
        print('==================')
        print('[harn] Scores per Config: %s' % testnameid)
        print('==================')
        #for cfgx in range(nConfig):
        #    print('[score] %s' % (cfgx2_lbl[cfgx]))
        #    for X in X_LIST:
        #        nLessX_ = nLessX_dict[int(X)][cfgx]
        #        print('        ' + rankscore_str(X, nLessX_, nQuery))
        print('\n[harn] ... sorted scores')
        for X in X_LIST:
            print('\n[harn] Sorted #ranks < %r scores' % (X))
            sortx = np.array(nLessX_dict[int(X)]).argsort()
            #frac_list = (nLessX_dict[int(X)] / cfgx2_nQuery)[:, None]
            #print('cfgx2_nQuery = %r' % (cfgx2_nQuery,))
            #print('frac_list = %r' % (frac_list,))
            #print('Pairwise Difference: ' + str(ut.safe_pdist(frac_list, metric=ut.absdiff)))
            for cfgx in sortx:
                nLessX_ = nLessX_dict[int(X)][cfgx]
                rankstr = rankscore_str(X, nLessX_, cfgx2_nQuery[cfgx], withlbl=False)
                print('[score] %s --- %s' % (rankstr, cfgx2_lbl[cfgx]))
    print_colscore(**kwargs)

    ut.argv_flag_dec(testres.print_percent_identification_success)(**kwargs)

    sumstrs = []
    sumstrs.append('++===========================')
    sumstrs.append('|| [cfg*] TestName: %s' % testnameid)
    sumstrs.append('||---------------------------')
    sumstrs.append(ut.joins('\n|| ', best_rankscore_summary))
    sumstrs.append('LL===========================')
    summary_str = '\n'.join(sumstrs)
    #print(summary_str)
    ut.colorprint(summary_str, 'blue')

    print('To enable all printouts add --print-all to the commandline')


def rankscore_str(thresh, nLess, total, withlbl=True):
    #helper to print rank scores of configs
    percent = 100 * nLess / total
    fmtsf = '%' + str(ut.num2_sigfig(total)) + 'd'
    if withlbl:
        fmtstr = ':#ranks < %d = ' + fmtsf + '/%d = (%.1f%%) (err=' + fmtsf + '/' + str(total) + ')'
        rankscore_str = fmtstr % (thresh, nLess, total, percent, (total - nLess))
    else:
        fmtstr = fmtsf + '/%d = (%.1f%%) (err=' + fmtsf + '/' + str(total) + ')'
        rankscore_str = fmtstr % (nLess, total, percent, (total - nLess))
    return rankscore_str


if __name__ == '__main__':
    """
    CommandLine:
        python -m ibeis.expt.experiment_printres
        python -m ibeis.expt.experiment_printres --allexamples
        python -m ibeis.expt.experiment_printres --allexamples --noface --nosrc
    """
    import multiprocessing
    multiprocessing.freeze_support()  # for win32
    import utool as ut  # NOQA
    ut.doctest_funcs()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
# UTool
import utool


# Inject utool functions
(print, print_, printDBG, rrr, profile) = utool.inject(
    __name__, '[preproc_rvecs]', DEBUG=False)


def add_rvecs_params_gen(ibs, nInput=None):
    pass


def generate_rvecs(vecs_list, words):
    pass
